/* cluster.c */
/* this program is an implementation of the clustering flow-chart 
   in the paper "Node-Tearing Nodal Analysis" Memo # ERL-M582 */

#include <stdio.h>

#include <X11/Xlib.h>    /* X-image */
#include <X11/Xutil.h>   /* X-image */
#include <X11/Xos.h>     /* X-image */
#include <math.h>        /* X-image */
#include <sys/file.h>    /* X-image */
#include "icon_bitmap"   /* X-image */

#define XDISPLAY       0
#define INFI  1073741824
#define YES            1
#define NO             0
#define FOREVER        1

#include "sets_cl.h"
#include "bdb_struct_chol_cluster.h"

/* define a grayscale bitmap that represents the sparsity pattern
   in the matrix 
    nonzero -> black=1
    zero    -> white=254
    fillin  -> gray =128 */

#define NONZERO        1 /* black */
#define ZERO         248 /* white */
#define FILLIN       124 /* gray  */

/* values for window_size in main, is window big enough to be useful */
#define SMALL        1   /* X-image */
#define OK           0   /* X-image */
#define MAX_COLORS 256   /* X-image */

/* The calculation of data clusters will be performed on the data pairs
   representation of the network.  This will simplify the network
   reordering - nodes just need to be renumbered.  After the nodes have
   been renumbered, then the A[][] matrix can be formed using a linked list
   and previously developed software can be used to perform symbolic
   factorization.  */

struct Node{int NodeNum;
            int NodeInUse;};

typedef struct AdjList{struct Node l;
                       struct Node r;
                       int fillin;} AdjList;

AdjList *adjlist;   /* adjlist[]  */
AdjList *adjlist1;  /* adjlist1[] */
int nadjlist;       /* number of adjacency pairs */
AdjList *adjlistF;  /* adjlistF[]  */
int nadjlistF;      /* number of adjacency pairs in the matrix with fillin */

typedef struct NodeListObj{int NodeNum;
                           int NodeInUse;} NodeList;

NodeList *nodelist; /* nodelist[] */

Set    *N2;           /* the set N2 */
SoSets *N1;           /* the set of sets of N1a... or the set of clusters */
SoSets *rN1;          /* the set of sets of N1a... recursively ordered set of clusters*/
SoSets *sN1;          /* the set of sets of N1a... ordered according to the super clusters */

/* The tableau is stored in stored as an array of structures -
   (1) the node number and (2) the degree of the node */

typedef struct Tabl{int NodeNum;
                    Set *AdjacentSet;
                    int ContourNum;} Tableau;

Tableau *tableau;   /* tableau[] */
int ntab;           /* number of elements in the tableau */

typedef struct Adjnds{int NodeInUse;
                      Set *AdjacentSet;} Adjnodes;
Adjnodes *AdjacentNodes; /* AdjacentNodes[] */

typedef struct SparseMatrixSet{Set *Row;
                               Set *Col;} SparseMatrixSet;
SparseMatrixSet *SparseMatrix; /* SparseMatrix[] */

typedef struct IndBlocks{Set *N1i;
                         int blockstart;
                         int blockend;} IndBlocks;
IndBlocks *IndBlk; /* IndBlk[] */

int n;               /* the number of columns/rows in A */
int ndisplay;        /* the number of columns/rows in the display */
int ncomb;           /* the number of pixels to be combined to condense the display */
int ntot;            /* the number of nonzeros in A */
int FillinCount;     /* a counter for the number of fillins */
int nOpsLastBlockFac;   /* number of operations in the last block */
int nOpsLastBlockFrBs;  /* number of operations in the last block */
int PrintN2;
double alpha;

int nmax;            /* maximum size of a cluster */
int nmin;            /* minimum size of a cluster */

/* define the structure to perform the pigeon-hole load balancing */

typedef struct LoadBalance{int    nOpsBlockFac;
                           int    nOpsBorderFac;
                           int    nOpsTotalFac;
                           int    nOpsBlockFrBs;
                           int    nOpsBorderFrBs;
                           int    nOpsTotalFrBs;
                           double nOpsTotal;
                           SoSets *SuperCluster;} LoadBalance;
LoadBalance *loadbal;   /* loadbal[] */

short **bitmap;      /* bitmap[][] display bitmap */ 

/* Display and screen are used as arguments to nearly every Xlib
 * routine, so it simplifies routine calls to declare them global.
 */
Display      *display;          /* X-image */
int          screen;            /* X-image */
XImage       *picture;          /* X-image */
unsigned int width, height;     /* X-image */
int    gargc;                   /* X-image */
char **gargv;                   /* X-image */
/*---------------------------------------------------------*/
Proc        *proc;       /* proc[]      */
Block       *blocks;     /* blocks[]    */
Diag        *diag;       /* diag[]      */
BlockRow    *blockrow;   /* blockrow[]  */
LowerBorder *lborder;    /* lborder[]   */
BorderRow   *borderrow;  /* borderrow[] */
/*---------------------------------------------------------*/
int            nblocks;    /* # of blocks            */
int            nblockrow;  /* dimension of blockrow  */
int            ndiag;      /* dimension of diag      */
int            nlborder;   /* dimension of lborder   */
int            nborderrow; /* dimension of borderrow */
/*---------------------------------------------------------*/
int    ln;    /* dimension of the last block */
int    lnnz;  /* number of non-zero in the last block */
LastBlock *lastbk;
/*---------------------------------------------------------*/
void main0();
void main1();
void main2();
void main3();
void main4();
void main5();
void main6();
void main7();
void main8();
void main9();
char *MakeOutputFilename();
void InputNetwork();
void ResetAdjList();
void IdClusters();
void SubClusterBlocks();
void ResetNodesInN1i();
void IdSubClusters();
char *MakeClusterName();
char *MakeSubClusterName();
void InitNodeList();
void InitAdjList();
void InitTableau();
int  StartNode();
void UpdateTableau();
int  SelectNextNodeInTableau();
void UpdateAdjacencyList();
void ResetAdjacencyList();
void ReorderNetworkClusters();
void MinDegreeReorder();
void NonZeroDensityInLastBlock();
void GenerateBitmap();
void GenerateBitmapF();
void PrintL();
void PrintLF();
void FillinBreakdown();
void FactorBBDM();
void Factor();
void PigeonHoleLoadBalance();
void BDBFormData();
void OutputBDBFormData();

double drand48();
void   srand48();

void Ximage();         /* X-image */
void set_gray_scale(); /* X-image */
void get_GC();         /* X-image */
int  load_data();      /* X-image */

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
main(argc,argv)
  int argc;
  char **argv;
{
  int i,j;
  int seed = 10770;
  int nprocsmin,nprocsmax;
  char filename[40];

  if (argc < 10)
   {
    printf("Usage: (1) matrix in adjacency list form \n");
    printf("       (2) nmax - maximum number of nodes in a cluster\n");
    printf("       (3) nmin - suggested minimum number of nodes in a cluster\n");
    printf("       (4) nprocsmin - min number of procs for load balancing\n");
    printf("       (5) nprocsmax - max number of procs for load balancing\n");
    printf("       (6) bitmap output name key (none)\n");
    printf("       (7) BDB-form matrix output name \n");
    printf("       (8) b vector output name \n");
    printf("       (9) list of edges in last block output name \n");
    printf("      (10) alpha - linear combination coefficient for load balancing \n");
    exit(1);
   } /* end if */

/* save the input string */
  gargc = argc;
  gargv = argv;

 printf("matrix in adjacency list form                     - argv[1]=%s\n",argv[1]);
 printf("maximum number of nodes in a cluster              - argv[2]=%s\n",argv[2]);
 printf("minimum number of nodes in a cluster              - argv[3]=%s\n",argv[3]);
 printf("minimum number of processors                      - argv[4]=%s\n",argv[4]);
 printf("maximum number of processors                      - argv[5]=%s\n",argv[5]);
 printf("bitmap output name key (none to suppress output)  - argv[6]=%s\n",argv[6]);
 printf("BDB-form matrix output name                       - argv[7]=%s\n",argv[7]);
 printf("b vector output name                              - argv[8]=%s\n",argv[8]);
 printf("list of edges in last block output name           - argv[9]=%s\n",argv[9]);
 printf("linear combination coefficient for load balancing - argv[10]=%s\n",argv[10]);

 sscanf(argv[2],"%d",&nmax);
 printf("maximum number of nodes in a cluster - nmax=%d \n",nmax);
 sscanf(argv[3],"%d",&nmin);
 printf("minimum number of nodes in a cluster - nmin=%d \n",nmin);
 sscanf(argv[4],"%d",&nprocsmin);
 printf("minimum number of processors - nprocsmin=%d \n",nprocsmin);
 sscanf(argv[5],"%d",&nprocsmax);
 printf("maximum number of processors - nprocsmax=%d \n",nprocsmax);
 sscanf(argv[10],"%lf",&alpha);
 printf("linear combination coefficient for load balancing - alpha=%lf \n",alpha);
 fflush(stdout);

/* initialize the SetElementPool */
  SetElementPool = NULL;
  SetElementPool = InitSet(SetElementPool,"SetElementPool");
  SetOfSetsElementPool = NULL;
  SetOfSetsElementPool = InitSetOfSets(SetOfSetsElementPool,"SetOfSetsElementPool");
  srand48(seed);
  AdjacentNodes = NULL;
  proc      = NULL;
  blocks    = NULL;
  diag      = NULL;
  blockrow  = NULL;
  lborder   = NULL;
  borderrow = NULL;
  lastbk    = NULL;
  PrintN2   = YES;

printf("InputNetwork \n");
fflush(stdout);
  InputNetwork(argv[1]); /* input pairs of data */

/* -------------------------------------------------------------------- */
/*
printf("%s data in original ordering \n",argv[6]);
  main0(argv[6]);
*/
/* -------------------------------------------------------------------- */
/*
printf("%s data after minimum degree reordering \n",argv[6]);
  main2(argv[6]);

printf("%s data after minimum degree reordering and Factorization \n",argv[6]);
  main3(argv[6]);
*/
/* -------------------------------------------------------------------- */

printf("%s data after clustering and minimum degree reordering \n",argv[6]);
  main4(argv[6]);
printf("%s data after clustering, minimum degree reordering, and Factorization \n",argv[6]);
  main5(argv[6]);
printf("%s data after superclustering \n",argv[6]);
  main6(argv[6],argv[7],argv[8],argv[9],nprocsmin,nprocsmax);

/* -------------------------------------------------------------------- */
/*
printf("%s data after subclustering and minimum degree reordering \n",argv[6]);
  main7(argv[6]);
printf("%s data after subclustering, minimum degree reordering, and Factorization \n",argv[6]);
  main8(argv[6]);
printf("%s data after superclustering \n",argv[6]);
  main9(argv[6],argv[7],argv[8],argv[9],nprocsmin,nprocsmax);
*/
/* -------------------------------------------------------------------- */
} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main0(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data in original ordering (main0) \n",namekey);
printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,0)); /* produce a bitmap of the fillin pattern */

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the fillin pattern */
} /* end main0() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main2(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after minimum degree reordering (main2) \n",namekey);
printf("ResetAdjList \n");
fflush(stdout);
  ResetAdjList();  /* reset the adjacency List to the initial state */

printf("MinDegreeReorder  \n");
fflush(stdout);
  MinDegreeReorder(); /* determine a minimum degree ordering of the matrix */

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,2)); /* produce a bitmap of the fillin pattern */

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the fillin pattern */
} /* end main2() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main3(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after minimum degree reordering and Factorization (main3) \n",namekey);
printf("GenerateBitmapF \n");
fflush(stdout);
  GenerateBitmapF(MakeOutputFilename(namekey,filename,3)); /* produce a bitmap of the fillin pattern */

printf("Factor \n");
fflush(stdout);
  Factor(); /* factor the matrix */

printf("PrintLF \n");
fflush(stdout);
  PrintLF("NiMo Data"); /* produce an X Window display of the fillin pattern */
} /* end main3() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main4(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after clustering and minimum degree reordering (main4)\n",namekey);
printf("ResetAdjList \n");
fflush(stdout);
  ResetAdjList();  /* reset the adjacency List to the initial state */

printf("IdClusters \n");
fflush(stdout);
  IdClusters();  /* identify clusters */

printf("ReorderNetworkClusters \n");
fflush(stdout);
  ReorderNetworkClusters(N1); /* produce block-bordered matrix */

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,4)); /* produce a bitmap of the fillin pattern */

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the fillin pattern */
} /* end main4() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main5(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after clustering, minimum degree reordering, and Factorization (main5)\n",namekey);

printf("Factor \n");
fflush(stdout);
  Factor(); /* factor the matrix */

printf("FactorBBDM \n");
fflush(stdout);
  FactorBBDM(); /* factorize the matrix */

printf("GenerateBitmapF \n");
fflush(stdout);
  GenerateBitmapF(MakeOutputFilename(namekey,filename,5)); /* produce a bitmap of the fillin pattern */

printf("FillinBreakdown \n");
fflush(stdout);
  FillinBreakdown();

printf("NonZeroDensityInLastBlock \n");
fflush(stdout);
  NonZeroDensityInLastBlock(); /* calculate the nonzero density in the last block */

printf("PrintLF \n");
fflush(stdout);
  PrintLF("NiMo Data"); /* produce an X Window display of the fillin pattern */
} /* end main5() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main6(namekey,FileNameA,FileNameb,FileNameN2,nprocsmin,nprocsmax)
  char *namekey,*FileNameA,*FileNameb,*FileNameN2;
  int nprocsmin,nprocsmax;
{
  char filename[40];
  int i,nprocs;

printf("%s data after superclustering (main6) \n",namekey);

  i = 0;
  for (nprocs=nprocsmin; nprocs<=nprocsmax; nprocs *= 2)
   {
printf("\n\nPigeonHoleLoadBalance  --  nprocs=%d\n",nprocs);
fflush(stdout);
 /* calculate the load balance for nprocs processors */
    PigeonHoleLoadBalance(N1,FileNameA,FileNameb,FileNameN2,namekey,nprocs);

   } /* end for nprocs */

} /* end main6() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main7(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after subclustering and minimum degree reordering (main7) \n",namekey);
printf("ResetAdjList \n");
fflush(stdout);
  ResetAdjList();  /* reset the adjacency List to the initial state */

printf("IdClusters \n");
fflush(stdout);
  IdClusters();  /* identify clusters */

printf("SubClusterBlocks \n");
fflush(stdout);
  SubClusterBlocks();  /* identify subclusters */

printf("ReorderNetworkClusters \n");
fflush(stdout);
  ReorderNetworkClusters(rN1); /* produce recursive block-bordered matrix */

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,7)); /* produce a bitmap of the fillin pattern */

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the fillin pattern */

} /* end main7() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main8(namekey)
  char *namekey;
{
  char filename[40];

printf("%s data after subclustering, minimum degree reordering, and Factorization (main8)\n",namekey);

printf("Factor \n");
fflush(stdout);
  Factor(); /* factorize the matrix */

printf("FactorBBDM \n");
fflush(stdout);
  FactorBBDM(); /* factorize the matrix */

printf("GenerateBitmapF \n");
fflush(stdout);
  GenerateBitmapF(MakeOutputFilename(namekey,filename,8)); /* produce a bitmap of the fillin pattern */

printf("FillinBreakdown \n");
fflush(stdout);
  FillinBreakdown();

printf("NonZeroDensityInLastBlock \n");
fflush(stdout);
  NonZeroDensityInLastBlock(); /* calculate the nonzero density in the last block */

printf("PrintL \n");
fflush(stdout);
  PrintLF("NiMo Data"); /* produce an X Window display of the fillin pattern */

} /* end main8() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main9(namekey,FileNameA,FileNameb,FileNameN2,nprocsmin,nprocsmax)
  char *namekey,*FileNameA,*FileNameb,*FileNameN2;
  int nprocsmin,nprocsmax;
{
  char filename[40];
  int i,nprocs;

printf("%s data after superclustering (main9) \n",namekey);

  i = 0;
  for (nprocs=nprocsmin; nprocs<=nprocsmax; nprocs *= 2)
   {
printf("PigeonHoleLoadBalance  --  nprocs=%d\n",nprocs);
fflush(stdout);
/* calculate the load balance for nprocs processors */
    PigeonHoleLoadBalance(rN1,FileNameA,FileNameb,FileNameN2,namekey,nprocs); 

   } /* end for nprocs */

} /* end main9() */



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeOutputFilename(key,filename,i)
  char *key,*filename;
  int i;
{
  int j,len;

  if (strcmp(key,"none") == 0)
   {
    strcpy(filename,key);
   } /* end if */
  else
   {
    len = strlen(key);
    for (j=0; j<len; j++)
      filename[j] = key[j];
    filename[j++] = '-';
    filename[j++] = (char) (((nmax%1000)/100) + 48);
    filename[j++] = (char) (((nmax%100)/10) + 48);
    filename[j++] = (char) ((nmax%10) + 48);
    filename[j++] = '-';
    filename[j++] = (char) (((nmin%1000)/100) + 48);
    filename[j++] = (char) (((nmin%100)/10) + 48);
    filename[j++] = (char) ((nmin%10) + 48);
    filename[j++] = '-';
    filename[j++] = (char) (((i%1000)/100) + 48);
    filename[j++] = (char) (((i%100)/10) + 48);
    filename[j++] = (char) ((i%10) + 48);
    filename[j++] = '\0';
   } /* end else */
/*
printf("Output filename for i=%d is |%s| \n",i,filename);
*/

  return filename;
} /* end MakeOutputFilename() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InputNetwork(filename)
  char *filename;
{
 int i,j,k,l,*histo,*hhisto;
 char *name;
 short *temp;
 FILE *fd;

 printf("read in the sparse matrix\n");
 fflush(stdout);

/* read in the sparse matrix */

printf("filename=|%s| \n",filename);
fflush(stdout);

 fd = fopen(filename,"r");

 fscanf(fd,"%d %d",&n,&nadjlist);

printf("n=|%d| -- nadjlist=|%d| \n",n,nadjlist);
fflush(stdout);

/* allocate memory for the adjacency list */
 adjlist = (AdjList *) calloc((nadjlist+1),sizeof(AdjList));
 if (adjlist == NULL)
  {
   printf("can't allocate memory for adjlist[] \n");
   exit(1);
  } /* end if */
 adjlist1 = (AdjList *) calloc((nadjlist+1),sizeof(AdjList));
 if (adjlist1 == NULL)
  {
   printf("can't allocate memory for adjlist1[] \n");
   exit(1);
  } /* end if */
/* allocate memory for the histogram */
 histo = (int *) calloc((n+1),sizeof(int));
 if (histo == NULL)
  {
   printf("can't allocate memory for histo[] \n");
   exit(1);
  } /* end if */
 for(i=0;i<=n;i++) histo[i]=0;
/* allocate memory for the histogram of histograms*/
 hhisto = (int *) calloc((n+1),sizeof(int));
 if (hhisto == NULL)
  {
   printf("can't allocate memory for hhisto[] \n");
   exit(1);
  } /* end if */
 for(i=0;i<=n;i++) hhisto[i]=0;

 for (k=1;k<=nadjlist;k++)
  {
   fscanf(fd,"%d  %d %d",&l,&i,&j);
/*
printf("  |%d| - (%d,%d)\n",k,i,j);
fflush(stdout);
*/
   adjlist[k].l.NodeNum = i;
   adjlist[k].r.NodeNum = j;
   adjlist1[k].l.NodeNum = i;
   adjlist1[k].r.NodeNum = j;

   adjlist[k].l.NodeInUse = YES;
   adjlist[k].r.NodeInUse = YES;
   adjlist1[k].l.NodeInUse = YES;
   adjlist1[k].r.NodeInUse = YES;

  histo[i]++; histo[j]++;
  }

 fclose(fd);

 for(i=1;i<=n;i++)
   {
    hhisto[histo[i]]++;
   } /* end for i */

 printf("\nHistogram of the number of nodes with i edges \n\n");
 for(i=1;i<=20;i++)
   {
    printf("%d --- \thhisto[%d]=%d \n",i,i,hhisto[i]);
    fflush(stdout);
   } /* end for i */

/* allocate memory for remining vectors */

  tableau = (Tableau *) calloc((n+1),sizeof(Tableau));
  if (tableau == NULL)
   {
    printf("can't allocate memory for tableau[] \n");
    exit(1);
   } /* end if */

  nodelist = (NodeList *) calloc((n+1),sizeof(NodeList));
  if (nodelist == NULL)
   {
    printf("can't allocate memory for nodelist[] \n");
    exit(1);
   } /* end if */

  SparseMatrix = (SparseMatrixSet *) calloc((n+1),sizeof(SparseMatrixSet));
  if (SparseMatrix == NULL)
   {
    printf("can't allocate memory for SparseMatrix[] \n");
    exit(1);
   } /* end if */

/* ititialize the SparseMatrix sets */
  for (i=1;i<=n;i++)
    {
     SparseMatrix[i].Row = NULL;
     name = (char *) calloc(10,sizeof(char));
     if (name == NULL)
      {
       printf("can't allocate memory for name[] in InputNetwork()\n");
       exit(1);
      } /* end if */
     name[0]  = 'R';
     name[1]  = 'o';
     name[2]  = 'w';
     name[3]  = '-';
     name[4]  = (char) (((i%100000)/10000) + 48);
     name[5]  = (char) (((i%10000)/1000) + 48);
     name[6]  = (char) (((i%1000)/100) + 48);
     name[7]  = (char) (((i%100)/10)  + 48);
     name[8]  = (char)  ((i%10)  + 48);
     name[9] = '\0';
     SparseMatrix[i].Row = InitSet(SparseMatrix[i].Row,name);
/*
PrintSet(SparseMatrix[i].Row);
*/
     SparseMatrix[i].Col = NULL;
     name = (char *) calloc(10,sizeof(char));
     if (name == NULL)
      {
       printf("can't allocate memory for name[] in InputNetwork()\n");
       exit(1);
      } /* end if */
     name[0]  = 'C';
     name[1]  = 'o';
     name[2]  = 'l';
     name[3]  = '-';
     name[4]  = (char) (((i%100000)/10000) + 48);
     name[5]  = (char) (((i%10000)/1000) + 48);
     name[6]  = (char) (((i%1000)/100) + 48);
     name[7]  = (char) (((i%100)/10)  + 48);
     name[8]  = (char)  ((i%10)  + 48);
     name[9] = '\0';
     SparseMatrix[i].Col = InitSet(SparseMatrix[i].Col,name);
/*
PrintSet(SparseMatrix[i].Col);
*/
    } /* end for i */

/* to produce a legible X display reduce the image to less than 1k X 1k */
  ncomb = (n/1024)+1;
  ndisplay = ((n-1)/ncomb)+1;
printf("ncomb=%d -- ndisplay=%d \n",ncomb,ndisplay);
  temp = (short *) calloc(((ndisplay+1)*(ndisplay+1)),sizeof(short));
  if (temp == NULL)
   {
    printf("can't allocate memory for temp when allocating bitmap[][] \n");
    exit(1);
   } /* end if */
  bitmap = (short **) calloc((ndisplay+1),sizeof(short *));
  if (bitmap == NULL)
   {
    printf("can't allocate memory for bitmap[][] \n");
    exit(1);
   } /* end if */
  for (i=0; i<(ndisplay+1); i++)
    bitmap[i] = temp + i*(ndisplay+1);

} /* end InputNetwork() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetAdjList()
{
  int i;

 for (i=1;i<=nadjlist;i++)
  {
   adjlist[i].l.NodeNum = adjlist1[i].l.NodeNum;
   adjlist[i].r.NodeNum = adjlist1[i].r.NodeNum;

   adjlist[i].l.NodeInUse = YES;
   adjlist[i].r.NodeInUse = YES;
  } /* end for i */

} /* end ResetAdjList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void IdClusters()
{
 int i,j,ncluster,nnodelist,locmin,node,nextnode,nN2;
 int MinConNum;
 char *name;
 Set *N1i,*N1i_1,*Temp0;
 SetElement *pres;

/* initialize the sets N1i and N2 and set of sets N1 */
  N2 = NULL;
  N2 = InitSet(N2,"BorderEquations");
  N1 = NULL;
  N1 = InitSetOfSets(N1,"Sets of Clusters");
/* initialize the node list that identifies nodes that remain in the
   adjacency list and the corresponding adjacent cut set information */
  InitNodeList();
  nnodelist = n;
  ncluster = 0;

/* look for clusters until there are insufficient nodes to produce
   a cluster of minimum size */
  while (nnodelist > 0)
   {
/* initialize the tableau */
    InitTableau();
    ntab = 0;
    ncluster++;
/* initialize variables in preparation to search for a cluster */
    name = MakeClusterName(ncluster);
    N1i = NULL;
    N1i = InitSet(N1i,name);
/* identify the node with the minimum adjacent cut set */
    nextnode = StartNode();
/* add to the tableau as long as the sizeof the tableau is less
   than the maximum cluster size and there are nodes
   remaining in the node list */
    while ( (ntab < nmax) && (nnodelist > 0) )
     {
      if (nextnode <= 0)
       {
        printf("(nextnode <= 0) so this is a dead end \n");
        printf("ntab=%d  --  nnodelist=%d  --  nextnode=%d\n",ntab,nnodelist,nextnode);
        fflush(stdout);
        break;
       } /* end if nextnode < 0 */
      ntab++;
      nnodelist--;
      if (ntab == 1)
        N1i_1 = NULL;
      else
        N1i_1 = tableau[ntab-1].AdjacentSet;
      UpdateTableau(nextnode,N1i_1);
      node = nextnode;
/* identify whether or not the tableau can be expanded any further */
      if (tableau[ntab].ContourNum == 0) /* this is the dead-end problem */
       {
/*
        printf("tableau[%d].ContourNum = 0 \n",ntab); 
*/
        locmin = ntab;
        break;
       } /* end if tableau[ntab].NodeDegree == 0 */	
/* identify whether or not the tableau is sufficiently large as to look
   for a cluster break */
      if (ntab == nmin)
       {
        locmin = ntab;
        MinConNum = n;
       } /* end if */
      else if (ntab <= (3*nmax/4))
       {
        if (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum)
         {
          locmin = ntab;
         } /* end if */
       } /* end else if */
      else
       {
        if ( (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum) &&
             (tableau[ntab].ContourNum <= MinConNum) )
         {
          locmin = ntab;
          MinConNum = tableau[ntab].ContourNum;
         } /* end if */
       } /* end else */
      nextnode = SelectNextNodeInTableau();

      UpdateAdjacencyList(node);
     } /* end while ntab < nmax */
/* implement dynamic contour cutting by adding those nodes to N2 
   (the set of coupling equations) remaining after identifying a cluster */
    if (ntab <= nmin) locmin = ntab;
    if (locmin < ntab)
     {
/*
      printf("\n>\n>locmin=%d is less than ntab=%d - so update nnodelist \n",locmin,ntab);
*/
      nnodelist += (ntab-locmin);
      for (i=(locmin+1); i<=ntab; i++)
       {
        ResetAdjacencyList(tableau[i].NodeNum);
       } /* end for i */
     } /* end if */
    Temp0 = tableau[locmin].AdjacentSet;
/*
    PrintSet(Temp0);
*/
    nN2 = N2->n;
    UnionOfSets(N2,Temp0);
    nnodelist -= (N2->n - nN2);
/*
    PrintSet(N2);
*/

/* produce the set N1i by collecting all elements in the tableau into
   a set */
    for (i=1; i<=locmin; i++)
     {
      AddElementToSet(N1i,tableau[i].NodeNum);
     } /* end for i */

/* update the adjacency list to reflect removing both
   the cluster and the dynamic contour cut nodes */
    pres = Temp0->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */
    pres = N1i->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */

/* reset nodelist[].NodeNum and nodelist[].NodeInUse */
    InitNodeList();

/* add the set N1i to N1 - the set of cluster sets */
    AddSetToSetOfSets(N1,N1i);
/*
    PrintSet(N1i);
    PrintSet(N2);
*/
   } /* end while nnodelist > 0 */

/* add remaining nodes to the set N2 */
  for (i=1; i<=n; i++)
   {
    if (nodelist[i].NodeInUse == YES)
     {
      AddElementToSet(N2,i);
     } /* end if */
   } /* end for i */
/*
  PrintSetOfSets(N1);
  PrintSet(N2);
*/
} /* end IdClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void SubClusterBlocks()
{
  int i,j;
  Set *N1i,*RN1i,*RN2;
  SetElement *element;
  SoSets *RN1,*TN1;
  SoSElement *soselN1,*soselRN1;

  TN1  = N1;

  N1  = NULL;
  N1  = InitSetOfSets(N1 ,TN1->Name);
  rN1 = NULL;
  rN1 = InitSetOfSets(rN1,TN1->Name);
/*
printf("SubClusterBlocks  --  set of clusters = |%s| \n",TN1->Name);
fflush(stdout);
*/
  soselN1 = TN1->First;
  while (soselN1 != NULL)
   {
    N1i = soselN1->set;
    AddSetToSetOfSets(N1,N1i);
/*
printf("SubClusterBlocks  --  N1i->Name=|%s| N1i->n=%d\n",N1i->Name,N1i->n);
fflush(stdout);
*/
    if ((N1i->n) > nmin)
     {
/* reset adjlist.x.NodeInUse to reflect only those nodes in the cluster */
      ResetNodesInN1i(N1i);
      InitNodeList();

/* initialize the set of sets RN1 and the set RN2 */
      RN2 = NULL;
      RN2 = InitSet(RN2,"ClusterBorderEquations");
      RN1 = NULL;
      RN1 = InitSetOfSets(RN1,N1i->Name);

/* generate the subclusters */
/*
printf("generate the subclusters for cluster = |%s| \n",N1i->Name);
fflush(stdout);
*/
      IdSubClusters(N1i,RN1,RN2);
/*
      PrintSet(N1i);
      PrintSetOfSets(RN1);
      PrintSet(RN2);
*/
/* reorder N1 to reflect the subclustering */
/* place the set RN2 into rN1 */
      AddSetToSetOfSets(rN1,RN2);
/* place the values in each subcluster into descending degree order */
      soselRN1 = RN1->First;
      while (soselRN1 != NULL)
       {
        RN1i = soselRN1->set;
/* place the set RN1i back into rN1i */
        AddSetToSetOfSets(rN1,RN1i);
        soselRN1 = soselRN1->Next;
       } /* end while */
     } /* end if */
    else
     {
/* place the set N1i into N1 */
/* place the values in N1i back into rN1 */
      AddSetToSetOfSets(rN1,N1i);
     } /* end else */

    soselN1 = soselN1->Next;
   } /* end while */

} /* end SubClusterBlocks() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetNodesInN1i(N1i)
  Set *N1i;
{
  int i;
  SetElement *element;

/* reset adjlist.x.NodeInUse to reflect only those nodes in the cluster */
/* set all nodes in edge conbinations to NO */
  for (i=1; i<=nadjlist; i++)
   {
    adjlist[i].l.NodeInUse = NO;
    adjlist[i].r.NodeInUse = NO;
   } /* end for i */
/* set all nodes in clusters to YES */
  element = N1i->First;
  while (element != NULL)
   {
    for (i=1; i<=nadjlist; i++)
     {
      if (adjlist[i].l.NodeNum == element->Element)
        adjlist[i].l.NodeInUse = YES;
      if (adjlist[i].r.NodeNum == element->Element)
        adjlist[i].r.NodeInUse = YES;
     } /* end for i */
    element = element->Next;
   } /* end while */
/* check to see that both nodes in an edge are in the cluster */
  for (i=1; i<=nadjlist; i++)
   {
    if ( (adjlist[i].l.NodeInUse !=YES) || (adjlist[i].r.NodeInUse !=YES) )
     {
      adjlist[i].l.NodeInUse = NO;
      adjlist[i].r.NodeInUse = NO;
     } /* end if */
   } /* end for i */
} /* end ResetNodesInN1i() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void IdSubClusters(N1i,RN1,RN2)
  Set *N1i,*RN2;
  SoSets *RN1;
{
 int i,j,ncluster,nnodelist,locmin,node,nextnode,nRN2;
 int MinConNum;
 char *name;
 Set *RN1i,*RN1i_1,*Temp0;
 SetElement *pres;

  nnodelist = N1i->n;
  ncluster = 0;

/* look for clusters */
  while (nnodelist > 0)
   {
/* initialize the tableau */
    InitTableau();
    ntab = 0;
    ncluster++;
/* initialize variables in preparation to search for a cluster */
    name = MakeSubClusterName(ncluster);
    RN1i = NULL;
    RN1i = InitSet(RN1i,name);
/* identify the node with the minimum adjacent cut set */
    nextnode = StartNode();
/* add to the tableau as long as the sizeof the tableau is less
   than the maximum cluster size and there are nodes
   remaining in the node list */
    while ( (ntab < nmin) && (nnodelist > 0) )
     {
      if (nextnode <= 0)
       {
        printf("(nextnode <= 0) so this is a dead end \n");
        printf("ntab=%d  --  nnodelist=%d  --  nextnode=%d\n",ntab,nnodelist,nextnode);
        fflush(stdout);
        break;
       } /* end if nextnode < 0 */
      ntab++;
      nnodelist--;
      if (ntab == 1)
        RN1i_1 = NULL;
      else
        RN1i_1 = tableau[ntab-1].AdjacentSet;
/*
printf("UpdateTableau  --  nextnode = %d\n",nextnode);
fflush(stdout);
*/
      UpdateTableau(nextnode,RN1i_1);
      node = nextnode;
/* identify whether or not the tableau can be expanded any further */
      if (tableau[ntab].ContourNum == 0) /* this is the dead-end problem */
       {
/*
        printf("tableau[%d].ContourNum = 0 \n",ntab);
*/
        locmin = ntab;
        break;
       } /* end if tableau[ntab].NodeDegree == 0 */
/* identify whether or not the tableau is sufficiently large as to look
   for a cluster break */
      if (ntab == 1) 
       {
        locmin = ntab;
        MinConNum = n;
       } /* end if */
      else if (ntab <= (nmin/2)) 
       {
        if (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum)
         {
          locmin = ntab;
         } /* end if */
       } /* end else if */
      else
       {
        if ( (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum) &&
             (tableau[ntab].ContourNum <= MinConNum) )
         {
          locmin = ntab;
          MinConNum = tableau[ntab].ContourNum;
         } /* end if */
       } /* end else */
      nextnode = SelectNextNodeInTableau();
/*
printf("nextnode = SelectNextNodeInTableau() = %d\n",nextnode);
fflush(stdout);
*/
      UpdateAdjacencyList(node);
     } /* end while ntab <= (nmax/8) */
/* implement dynamic contour cutting by adding those nodes to RN2
   (the set of coupling equations) remaining after identifying a cluster */
    if (ntab == 1) locmin = ntab;
    if (locmin < ntab)
     {
/*
      printf("\n>\n>locmin=%d is less than ntab=%d - so update nnodelist \n",locmin,ntab);
*/
      nnodelist += (ntab-locmin);
      for (i=(locmin+1); i<=ntab; i++)
       {
        ResetAdjacencyList(tableau[i].NodeNum);
       } /* end for i */
     } /* end if */
    Temp0 = tableau[locmin].AdjacentSet;
/*
    PrintSet(Temp0);
*/
    nRN2 = RN2->n;
    UnionOfSets(RN2,Temp0);
    nnodelist -= (RN2->n - nRN2);
/*
    PrintSet(RN2);
*/

/* produce the set RN1i by collecting all elements in the tableau into
   a set */
    for (i=1; i<=locmin; i++)
     {
      AddElementToSet(RN1i,tableau[i].NodeNum);
     } /* end for i */

/* update the adjacency list to reflect removing both
   the cluster and the dynamic contour cut nodes */
    pres = Temp0->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */
    pres = RN1i->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */

/* reset nodelist[].NodeNum and nodelist[].NodeInUse */
    InitNodeList();

/* add the set RN1i to RN1 - the set of cluster sets */
    AddSetToSetOfSets(RN1,RN1i);
   } /* end while nnodelist > 0 */

/* add remaining nodes to the set RN2 */
  for (i=1; i<=n; i++)
   {
    if (nodelist[i].NodeInUse == YES)
     {
      AddElementToSet(RN2,i);
     } /* end if */
   } /* end for i */
/*
  PrintSetOfSets(RN1);
  PrintSet(RN2);
*/
} /* end IdSubClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeClusterName(ncluster)
  int  ncluster;
{
  char *name;
  char *string = "Cluster";
  int i;

  name = (char *) calloc(12,sizeof(char));
  if (name == NULL)
   {
    printf("can't allocate memory for name[] in MakeClusterName()\n");
    exit(1);
   } /* end if */

  for (i=0; i<7; i++)
   name[i] = string[i];

  name[7]  = '-';
  name[8]  = (char) (((ncluster%1000)/100) + 48);
  name[9]  = (char) (((ncluster%100)/10)  + 48);
  name[10] = (char) ((ncluster%10)  + 48);
  name[11] = '\0';

} /* end MakeClusterName(); */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeSubClusterName(ncluster)
  int  ncluster;
{
  char *name;
  char *string = "SubCluster";
  int i;

  name = (char *) calloc(15,sizeof(char));
  if (name == NULL)
   {
    printf("can't allocate memory for name[] in MakeSubClusterName()\n");
    exit(1);
   } /* end if */

  for (i=0; i<10; i++)
   name[i] = string[i];

  name[10]  = '-';
  name[11]  = (char) (((ncluster%1000)/100) + 48);
  name[12]  = (char) (((ncluster%100)/10)  + 48);
  name[13] = (char) ((ncluster%10)  + 48);
  name[14] = '\0';

} /* end MakeClusterName(); */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InitNodeList()
{
  int i;

/* initialize nodelist[].NodeNum & nodelist[].NodeInUse */
  for (i=1;i<=n;i++)
   {
    nodelist[i].NodeNum = 0;
    nodelist[i].NodeInUse = NO;
   } /* end for i */

/* calculate the number of connections at each node */

  for (i=1;i<=nadjlist;i++)
   {
    if (adjlist[i].l.NodeInUse == YES)
     {
      (nodelist[ adjlist[i].l.NodeNum ].NodeNum)++;
      nodelist[ adjlist[i].l.NodeNum ].NodeInUse = YES;
     } /* end if */
    if (adjlist[i].r.NodeInUse == YES)
     {
      (nodelist[ adjlist[i].r.NodeNum ].NodeNum)++;
      nodelist[ adjlist[i].r.NodeNum ].NodeInUse = YES;
     } /* end if */
   } /* end for i */

/*
  for (i=1;i<=n;i++)
   {
printf("nodelist[%d].NodeNum=%d  --  nodelist[%d].NodeInUse=%d\n",i,nodelist[i].NodeNum,i,nodelist[i].NodeInUse);
fflush(stdout);
   } /* end for i */
} /* end for InitNodeList() */



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InitAdjList()
{
  int i;

/* initialize AdjacentNodes[].NodeInUse & AdjacentNodes[].AdjacentSet */
  for (i=1;i<=n;i++)
   {
    AdjacentNodes[i].NodeInUse = NO;
    AdjacentNodes[i].AdjacentSet = NULL;
    AdjacentNodes[i].AdjacentSet = InitSet(AdjacentNodes[i].AdjacentSet,"Adjnodes");
   } /* end for i */

/* determine the set of adjacent nodes at each node */

  for (i=1;i<=nadjlist;i++)
   {
    if ((adjlist[i].l.NodeInUse == YES) && (adjlist[i].r.NodeInUse == YES))
     {
      AdjacentNodes[ adjlist[i].l.NodeNum ].NodeInUse = YES;
      AdjacentNodes[ adjlist[i].r.NodeNum ].NodeInUse = YES;
      AddElementToSet(AdjacentNodes[ adjlist[i].l.NodeNum ].AdjacentSet,adjlist[i].r.NodeNum);
      AddElementToSet(AdjacentNodes[ adjlist[i].r.NodeNum ].AdjacentSet,adjlist[i].l.NodeNum);
     } /* end if */
   } /* end for i */
} /* end for InitAdjList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InitTableau()
{
  int i;

  for (i=1; i<=n; i++)
   {
    tableau[i].NodeNum = 0;
    tableau[i].ContourNum = 0;
    if (tableau[i].AdjacentSet != NULL)
      DeleteSet(tableau[i].AdjacentSet);
   } /* end for i */
} /* end InitTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int StartNode()
{
 int i,MinDeg,MinDegNode;

/* find the minimum degree node */

  MinDeg = n;
  MinDegNode = 0;

  for (i=1;i<=n;i++)
   {
    if ( (nodelist[i].NodeNum < MinDeg) && (nodelist[i].NodeInUse == YES) )
     {
      MinDeg = nodelist[i].NodeNum;
      MinDegNode = i;
     } /* end if */
   } /* end for i */

if (MinDegNode == 0)
 {
  printf("In StartNode() -- MinDegNode = 0 \n");
  fflush(stdout);
 } /* end if */

 return MinDegNode;

} /* end StartNode() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void UpdateTableau(nextnode,N1i_1)
  int nextnode;
  Set *N1i_1;
{
  int i;
  Set *adjset;
  SetElement *pres;

  tableau[ntab].NodeNum = nextnode;

/* generate the set of adjacent nodes to this node */
/* first - find all node adjacent to this node */
  adjset = NULL;
  adjset = InitSet(adjset,"Adjacent Nodes");
  for (i=1; i<=nadjlist; i++)
   {
/*
printf("UpdateTableau() -- i=%d \n",i);
fflush(stdout);
*/
    if( (adjlist[i].l.NodeNum == nextnode) && 
        (adjlist[i].l.NodeInUse == YES)    &&
        (adjlist[i].r.NodeInUse == YES) )
     {
/*
printf("AddElementToSet(adjset,%d) \n",adjlist[i].r.NodeNum);
fflush(stdout);
*/
      AddElementToSet(adjset,adjlist[i].r.NodeNum);
     } /* end if */
    if( (adjlist[i].r.NodeNum == nextnode) && 
        (adjlist[i].r.NodeInUse == YES)    &&
        (adjlist[i].l.NodeInUse == YES) )
     {
/*
printf("AddElementToSet(adjset,%d) \n",adjlist[i].l.NodeNum);
fflush(stdout);
*/
      AddElementToSet(adjset,adjlist[i].l.NodeNum);
     } /* end if */
   } /* end for i */

  if (ntab > 1)
   {
/*
printf("print out the adjacent set from the previous tableau entry \n");
PrintSet(N1i_1);
printf("print out the partial adjacent set from this tableau entry \n");
PrintSet(adjset);
*/
/* second - take the union of the set of adjacent nodes with the set of 
   adjacent nodes at the previous node in the adjacent set */
    UnionOfSets(adjset,N1i_1);
/* third - remove this node from the adjacent set */
    DeleteElementFromSet(adjset,nextnode);
/*
printf("print out the updated adjacent set from this tableau entry \n");
fflush(stdout);
PrintSet(adjset);
*/
   } /* end if (ntab > 1) */

  tableau[ntab].ContourNum = adjset->n;

/* print the header */
/*
  printf("%d - NodeNum=%d - AdjSet={",
          ntab,tableau[ntab].NodeNum);
  pres = adjset->First;
  while (pres != NULL)
   {
    printf("(%d,%d)",pres->Element,nodelist[pres->Element].NodeNum);
    pres = pres->Next;
   } /* end while */
/*
  printf("} - ContourNum=%d \n",tableau[ntab].ContourNum);
*/
  tableau[ntab].AdjacentSet = adjset;
  
} /* end UpdateTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int SelectNextNodeInTableau()
{
  int i,node,MinDeg,MinDegNode;
  SetElement *pres;
  Set *set;

/* check each element in the set to determine which element has the
   fewest connecting nodes given that this node hasn`t already been
   chosen.  If there are ties chose the first value encountered */

  MinDeg = n;
  MinDegNode = 0;

  set = tableau[ntab].AdjacentSet;
  pres = set->First;
  while (pres != NULL)
   {
    node = pres->Element;
    if ( (nodelist[node].NodeNum < MinDeg) && (nodelist[node].NodeInUse == YES) )
     {
      MinDeg = nodelist[node].NodeNum;
      MinDegNode = node;
     } /* end if */
    pres = pres->Next;
   } /* end while pres != NULL */

  return MinDegNode;

} /* end SelectNextNodeInTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetAdjacencyList(node)
  int node;
{
  int i;

/* mark this node as again accessible for further consideration */
  for (i=1; i<=nadjlist; i++)
   {
    if (node == adjlist[i].l.NodeNum)
      adjlist[i].l.NodeInUse = YES;
    if (node == adjlist[i].r.NodeNum)
      adjlist[i].r.NodeInUse = YES;
   } /* end for i */

} /* end ResetAdjacencyList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void UpdateAdjacencyList(node)
  int node;
{
  int i;

/* mark this node as not accessible for further consideration */
  for (i=1; i<=nadjlist; i++)
   {
    if (node == adjlist[i].l.NodeNum)
      adjlist[i].l.NodeInUse = NO;
    if (node == adjlist[i].r.NodeNum)
      adjlist[i].r.NodeInUse = NO;
   } /* end for i */

} /* end UpdateAdjacencyList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ReorderNetworkClusters(TN1)
  SoSets *TN1;
{
  int i,j,k,here;
  int MinDeg,MinDegLoc,nN1i,nN2,node;
  int *ReorderLst;  /* ReorderLst[] */
  Set *N1i;
  SetElement *element,*pres;
  SoSElement *soselement;

printf("ReorderNetworkClusters --- TN1->n=|%d| TN1->Name=|%s| \n",
                                   TN1->n,TN1->Name);
printf("\t\tn=%d - nadjlist=%d \n",n,nadjlist);
fflush(stdout);

  if (AdjacentNodes != NULL) free(AdjacentNodes);
  AdjacentNodes = (Adjnodes *) calloc((n+1),sizeof(Adjnodes));
  if (AdjacentNodes == NULL)
   {
    printf("can't allocate memeory for AdjacentNodes[] \n");
    exit(1);
   } /* end if */

  ReorderLst = (int *) calloc((n+1),sizeof(int));
  if (ReorderLst == NULL)
   {
    printf("can't allocate memeory for ReorderLst[] \n");
    exit(1);
   } /* end if */

/* reset adjlist.x.NodeInUse to YES */
  
  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeInUse = YES;
    adjlist[i].r.NodeInUse = YES;
   } /* end for i */

/* initialize the sets of adjacent nodes to all nodes*/
  InitAdjList();

/* determine a minimum degree ordering within clusters */
  i=0;
  soselement = TN1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    nN1i = N1i->n;
/*
printf("Cluster set -- |%s| nN1i = %d \n",N1i->Name,nN1i);
fflush(stdout);
*/
/*
PrintSet(N1i);
fflush(stdout);
*/
    for (j=1; j<=nN1i; j++)
     {
/*
printf("Within loop -- j = %d --------\n",j);
fflush(stdout);
*/
      element = N1i->First;
      MinDeg = n;
      MinDegLoc = 0;
      while (element != NULL)
       {
        node = element->Element;
/*
printf("Within loop -- j = %d --------node = %d\n",j,node);
fflush(stdout);
*/
        if ( (AdjacentNodes[node].AdjacentSet->n < MinDeg) && (AdjacentNodes[node].NodeInUse == YES) )
         {
          MinDeg = AdjacentNodes[node].AdjacentSet->n;
          MinDegLoc = node;
         } /* end if */
        element = element->Next;
/*
printf("Within loop -- j = %d --- element = %d\n",j,element);
fflush(stdout);
*/
       } /* end while element != NULL */
/*
printf("%s -- j = %d - MinDegLoc = %d - AdjacentNodes[%d].AdjacentList->n = %d \n",
           N1i->Name,j,MinDegLoc,MinDegLoc,AdjacentNodes[MinDegLoc].AdjacentSet->n);
fflush(stdout);
*/
      pres = AdjacentNodes[MinDegLoc].AdjacentSet->First;
      while(pres != NULL)
       {
        node = pres->Element;
        UnionOfSets(AdjacentNodes[node].AdjacentSet,AdjacentNodes[MinDegLoc].AdjacentSet);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,node);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,MinDegLoc);

        pres = pres->Next;
       } /* end while pres != NULL */

      AdjacentNodes[MinDegLoc].NodeInUse = NO;
      ReorderLst[MinDegLoc] = ++i;
/*
printf("%s -- ReorderLst[%d] = %d AdjacentNodes[%d].AdjacentList->n = %d \n",
         N1i->Name,MinDegLoc,ReorderLst[MinDegLoc],MinDegLoc,AdjacentNodes[MinDegLoc].AdjacentSet->n);
fflush(stdout);
*/
     } /* end for j */
    soselement = soselement->Next;
   } /* end while soselement != NULL */

/* determine a minimum degree ordering within the border */
/*
printf("determine a minimum degree ordering within the border \n");
*/
  nN2 = N2->n;
/*
printf("Border set -- |%s| nN2 - %d \n",N2->Name,nN2);
*/
/*
PrintSet(N2);
*/
  for (j=1; j<=nN2; j++)
   {
    element = N2->First;
    MinDeg = n;
    MinDegLoc = 0;
    while (element != NULL)
     {
      node = element->Element;
      if ( (AdjacentNodes[node].AdjacentSet->n < MinDeg) && (AdjacentNodes[node].NodeInUse == YES) )
       {
        MinDeg = AdjacentNodes[node].AdjacentSet->n;
        MinDegLoc = node;
       } /* end if */
      element = element->Next;
     } /* end while element != NULL */

      pres = AdjacentNodes[MinDegLoc].AdjacentSet->First;
      while(pres != NULL)
       {
        node = pres->Element;
        UnionOfSets(AdjacentNodes[node].AdjacentSet,AdjacentNodes[MinDegLoc].AdjacentSet);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,node);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,MinDegLoc);

        pres = pres->Next;
       } /* end while pres != NULL */

    AdjacentNodes[MinDegLoc].NodeInUse = NO;
    ReorderLst[MinDegLoc] = ++i;
/*
printf("%s -- ReorderLst[%d] = %d nodelist[%d].NodeNum= %d \n",
         N2->Name,MinDegLoc,ReorderLst[MinDegLoc],MinDegLoc,nodelist[MinDegLoc].NodeNum);
*/
   } /* end for j */

/* reset adjlist.x.NodeInUse to YES */
  for (j=1;j<=nadjlist;j++)
   {
    adjlist[j].l.NodeInUse = YES;
    adjlist[j].r.NodeInUse = YES;
   } /* end for j */

/* calculate the number of fillin */
  j = 0;
  for (i=1;i<=n;i++)
    {
     j += AdjacentNodes[i].AdjacentSet->n;
    } /* end for i */

  printf("The number of fillin from minimum-degree ordering ---\n");
  printf("\ttotal number of nodes = %d --- number of fillin = %d \n",j,(j-nadjlist));
  FillinCount = (j-nadjlist);

/* set up the data structure to store the edges with those edges created by fillin */
  nadjlistF = j;

/* allocate memory for the adjacency list */
 if (adjlistF != NULL) free(adjlistF);
 adjlistF = (AdjList *) calloc((nadjlistF+1),sizeof(AdjList));
 if (adjlistF == NULL)
  {
   printf("can't allocate memory for adjlistF[] \n");
   exit(1);
  } /* end if */

/* store edges including fillin */
  j = 0;
  for (i=1;i<=n;i++)
    {
     pres = AdjacentNodes[i].AdjacentSet->First;
     while(pres != NULL)
      {
       j++;
       node = pres->Element;
       adjlistF[j].r.NodeNum = i;
       adjlistF[j].l.NodeNum = node;
 
       adjlistF[j].fillin = YES;
       for (k=1;k<=nadjlist;k++)
	 {
          if( ( (adjlist[k].l.NodeNum == i) && (adjlist[k].r.NodeNum == node) ) ||
              ( (adjlist[k].r.NodeNum == i) && (adjlist[k].l.NodeNum == node) ) )
           adjlistF[j].fillin = NO;
	 } /* end for k */
            
       pres = pres->Next;
      } /* end while pres != NULL */
    } /* end for i */

/* reset adjlistF.x.NodeInUse to YES */
  for (j=1;j<=nadjlistF;j++)
   {
    adjlistF[j].l.NodeInUse = YES;
    adjlistF[j].r.NodeInUse = YES;
/*
printf("\t\t%d - (%d,%d) \n",j,adjlistF[j].l.NodeNum,adjlistF[j].r.NodeNum);
fflush(stdout);
    for (i=1;i<=nadjlistF;i++) {
       if ( (i!=j) && 
            (  ( (adjlistF[j].l.NodeNum == adjlistF[i].l.NodeNum) && 
                 (adjlistF[j].r.NodeNum == adjlistF[i].r.NodeNum) ) ||
               ( (adjlistF[j].l.NodeNum == adjlistF[i].r.NodeNum) && 
                 (adjlistF[j].r.NodeNum == adjlistF[i].l.NodeNum) ) ) )
	 {
           printf("Multiple nodes in adjlistF[].x.NodeNum --- i=%d - edge (%d,%d) - j=%d - edge (%d,%d) \n",
                   i,adjlistF[i].l.NodeNum,adjlistF[i].r.NodeNum,
                   j,adjlistF[j].l.NodeNum,adjlistF[j].r.NodeNum);
           fflush(stdout); } } 
*/
   } /* end for j */

  for (j=1;j<=n;j++)
   {
    here = NO;
    soselement = TN1->First;
    while (soselement != NULL)
     {
      N1i = soselement->set;
      if (ElementOfSet(N1i,j) == YES)
       {
        here = YES;
        break;
       } /* end if */
      soselement = soselement->Next;
     } /* end while soselement != NULL */
    if (here == NO)
     {
      if (ElementOfSet(N2,j) == NO)
        printf("** -- row %d is neither in a cluster nor in the border \n",j);
     } /* end if (here == NO) */
   } /* end for j */

/* exchange the node numbers in the adjacency list to reorder the matrix */
/*
printf("reorder the matrix \n");
*/
  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeNum = ReorderLst[adjlist[i].l.NodeNum];
    adjlist[i].r.NodeNum = ReorderLst[adjlist[i].r.NodeNum];
   } /* end for i */ 
  for (i=1;i<=nadjlistF;i++)
   {
    adjlistF[i].l.NodeNum = ReorderLst[adjlistF[i].l.NodeNum];
    adjlistF[i].r.NodeNum = ReorderLst[adjlistF[i].r.NodeNum];
   } /* end for i */

/* load the data into the SparseMatrix sets */
  for (i=1;i<=n;i++)
   {
    AddElementToSet(SparseMatrix[i].Col,i);
   } /* end for i */
  for (i=1;i<=nadjlistF;i++)
   {
    if (adjlistF[i].l.NodeNum > adjlistF[i].r.NodeNum)
      {
       AddElementToSet(SparseMatrix[adjlistF[i].l.NodeNum].Row,adjlistF[i].r.NodeNum);
       AddElementToSet(SparseMatrix[adjlistF[i].r.NodeNum].Col,adjlistF[i].l.NodeNum);
      } /* end if */
    else
      {
       AddElementToSet(SparseMatrix[adjlistF[i].r.NodeNum].Row,adjlistF[i].l.NodeNum);
       AddElementToSet(SparseMatrix[adjlistF[i].l.NodeNum].Col,adjlistF[i].r.NodeNum);
      } /* end else */
   } /* end for i */

/* free the memory */
  for (i=1;i<=n;i++)
    {
     DeleteSet(AdjacentNodes[i].AdjacentSet);
    } /* end for i */
  free(AdjacentNodes);
  free(ReorderLst);

} /* end ReorderNetworkClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void MinDegreeReorder()
{
  int i,j,k,here;
  int MinDeg,MinDegLoc,nN1i,nN2,node;
  int *ReorderLst;         /* ReorderLst[] */
  SetElement *pres;

  AdjacentNodes = (Adjnodes *) calloc((n+1),sizeof(Adjnodes));
  if (AdjacentNodes == NULL)
   {
    printf("can't allocate memeory for AdjacentNodes[] \n");
    exit(1);
   } /* end if */

  ReorderLst = (int *) calloc((n+1),sizeof(int));
  if (ReorderLst == NULL)
   {
    printf("can't allocate memeory for ReorderLst[] \n");
    exit(1);
   } /* end if */

/* reset adjlist.x.NodeInUse to YES */

  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeInUse = YES;
    adjlist[i].r.NodeInUse = YES;
   } /* end for i */

/* initialize the sets of adjacent nodes to all nodes*/
  InitAdjList();

/* determine a minimum degree ordering */

  for (i=1;i<=n;i++)
   {
    MinDeg = n;
    MinDegLoc = 0;

    for (j=1; j<=n; j++)
     {
      if ( (AdjacentNodes[j].AdjacentSet->n < MinDeg) && (AdjacentNodes[j].NodeInUse == YES) )
       {
        MinDeg = AdjacentNodes[j].AdjacentSet->n;
        MinDegLoc = j;
       } /* end if */
     } /* end for j */

     pres = AdjacentNodes[MinDegLoc].AdjacentSet->First;
     while(pres != NULL)
       {
        node = pres->Element;
        UnionOfSets(AdjacentNodes[node].AdjacentSet,AdjacentNodes[MinDegLoc].AdjacentSet);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,node);
        DeleteElementFromSet(AdjacentNodes[node].AdjacentSet,MinDegLoc);

        pres = pres->Next;
       } /* end while pres != NULL */

     AdjacentNodes[MinDegLoc].NodeInUse = NO;
     ReorderLst[MinDegLoc] = i;

/*
printf("Entire Matrix -- ReorderLst[%d] = %d nodelist[%d].NodeNum= %d \n",
         MinDegLoc,ReorderLst[MinDegLoc],MinDegLoc,nodelist[MinDegLoc].NodeNum);
*/
   } /* end for i */

/* calculate the number of fillin */
  j = 0;
  for (i=1;i<=n;i++)
    {
     j += AdjacentNodes[i].AdjacentSet->n;
    } /* end for i */

  printf("Estimates of the number of fillin from minimum-degree ordering \n");
  printf("\ttotal number of nodes = %d --- number of fillin = %d \n",j,(j-nadjlist));
  FillinCount = (j-nadjlist);

/* set up the data structure to store the edges with those edges created by fillin */
  nadjlistF = j;

/* allocate memory for the adjacency list */
 adjlistF = (AdjList *) calloc((nadjlistF+1),sizeof(AdjList));
 if (adjlistF == NULL)
  {
   printf("can't allocate memory for adjlistF[] \n");
   exit(1);
  } /* end if */

printf("store edges including fillin \n");
fflush(stdout);

/* store edges including fillin */
  j = 0;
  for (i=1;i<=n;i++)
    {
     pres = AdjacentNodes[i].AdjacentSet->First;
     while(pres != NULL)
      {
       j++;
       node = pres->Element;
       adjlistF[j].r.NodeNum = i;
       adjlistF[j].l.NodeNum = node;
 
       adjlistF[j].fillin = YES;
       for (k=1;k<=nadjlist;k++)
	 {
          if( ( (adjlist[k].l.NodeNum == i) && (adjlist[k].r.NodeNum == node) ) ||
              ( (adjlist[k].r.NodeNum == i) && (adjlist[k].l.NodeNum == node) ) )
           adjlistF[j].fillin = NO;
	 } /* end for k */
      
       pres = pres->Next;
      } /* end while pres != NULL */
    } /* end for i */

printf("reset adjlist.x.NodeInUse to YES \n");
fflush(stdout);

/* reset adjlist.x.NodeInUse to YES */
  for (j=1;j<=nadjlist;j++)
   {
    adjlist[j].l.NodeInUse = YES;
    adjlist[j].r.NodeInUse = YES;
   } /* end for j */

/* reset adjlistF.x.NodeInUse to YES */
  for (j=1;j<=nadjlistF;j++)
   {
    adjlistF[j].l.NodeInUse = YES;
    adjlistF[j].r.NodeInUse = YES;
   } /* end for j */

/* exchange the node numbers in the adjacency list to reorder the matrix */

printf("reorder the matrix \n");
fflush(stdout);

  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeNum = ReorderLst[adjlist[i].l.NodeNum];
    adjlist[i].r.NodeNum = ReorderLst[adjlist[i].r.NodeNum];
   } /* end for i */
  for (i=1;i<=nadjlistF;i++)
   {
    adjlistF[i].l.NodeNum = ReorderLst[adjlistF[i].l.NodeNum];
    adjlistF[i].r.NodeNum = ReorderLst[adjlistF[i].r.NodeNum];
   } /* end for i */

printf("load the data into the SparseMatrix sets \n");
fflush(stdout);

/* load the data into the SparseMatrix sets */
  for (i=1;i<=n;i++)
   {
    AddElementToSet(SparseMatrix[i].Col,i);
   } /* end for i */
  for (i=1;i<=nadjlistF;i++)
   {
    if (adjlistF[i].l.NodeNum > adjlistF[i].r.NodeNum)
      {
       AddElementToSet(SparseMatrix[adjlistF[i].l.NodeNum].Row,adjlistF[i].r.NodeNum);
       AddElementToSet(SparseMatrix[adjlistF[i].r.NodeNum].Col,adjlistF[i].l.NodeNum);
      } /* end if */
    else
      {
       AddElementToSet(SparseMatrix[adjlistF[i].r.NodeNum].Row,adjlistF[i].l.NodeNum);
       AddElementToSet(SparseMatrix[adjlistF[i].l.NodeNum].Col,adjlistF[i].r.NodeNum);
      } /* end else */
   } /* end for i */

/* free the memory */
  for (i=1;i<=n;i++)
    {
     DeleteSet(AdjacentNodes[i].AdjacentSet);
    } /* end for i */
  free(AdjacentNodes);
  free(ReorderLst);

} /* end MinDegreeReorder() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void NonZeroDensityInLastBlock()
{
  int i,j,lbz,lbnonz,lbfill;
  SetElement *pres;

  lbz = lbnonz = lbfill = 0;
  for (i=1; i<=nadjlistF; i++)
   {
    if ( (adjlistF[i].l.NodeNum > (n-(N2->n)) ) && (adjlistF[i].r.NodeNum > (n-(N2->n)) ) )
     {
      if (adjlistF[i].fillin == NO) lbnonz++;
      if (adjlistF[i].fillin == YES)  lbfill++;
     } /* end if */
   } /* end for i */

  lbz = ((N2->n) * ((N2->n)-1) / 2) - (lbnonz + lbfill);

  printf("The number of equations in the borders/last block = %d\n",N2->n);
  printf("The number of nonzeros in the last block          = %d\n",lbnonz);
  printf("The number of fillin in the last block            = %d\n",lbfill);
  printf("The number of zeros in the last block             = %d\n",lbz);
  printf("The percentage of nonzero & fillin                = %f\n",
           ( (double) ((lbnonz+lbfill) * 100) / (double) ( (((N2->n)-1)*(N2->n) / 2) ) ));
    
} /* NonZeroDensityInLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void GenerateBitmap(filename)
  char *filename;
{
 int i,j,k,ipixel;
 int idisplay,jdisplay;
 int PointerL;
 int Nonzero;
 int Fillin;

 FILE *fd;

  ntot=0;
/*
printf("\nIn GenerateBitmap() - n=%d \n",n);
*/

/* initialize the display bitmap to all white */
  for (i=1;i<=ndisplay;i++)
   for (j=1;j<=ndisplay;j++)
    bitmap[j][i] = ZERO;

/* enter the diagonal */
 for (k=1; k<=n; k++) /* for the diagonal */
  {
   idisplay = ((k-1)/ncomb)+1;
   jdisplay = ((k-1)/ncomb)+1;
   ntot++;
   bitmap[idisplay][jdisplay]=NONZERO;
  } /* end for k */

 for (k=1; k<=nadjlist; k++) /* for all edges */
  {
   i = adjlist[k].l.NodeNum;
   j = adjlist[k].r.NodeNum;
   idisplay = ((i-1)/ncomb)+1;
   jdisplay = ((j-1)/ncomb)+1;
   ntot += 2;
   bitmap[jdisplay][idisplay]=NONZERO;
   bitmap[idisplay][jdisplay]=NONZERO;
 } /* end for k */

printf("\nIn GenerateBitmap() - filename = |%s| \n",filename);
fflush(stdout);

  if (strcmp(filename,"none") != 0)
   {
    fd = fopen(filename, "w");

printf("Open |%s| for output of the Bitmap\n",filename);
fflush(stdout);

    for (i=1;i<=ndisplay;i++)
     for (j=1;j<=ndisplay;j++)
      {
       ipixel = bitmap[j][i];
       if( (ipixel < 0) || (ipixel > 255) )
        {
printf("error in writting bitmap[%d][%d]=|%d| to %s\n",j,i,ipixel);
fflush(stdout);
         ipixel = 0;
        } /* end if */
       fputc(ipixel,fd);
      }

    fclose(fd);

printf("Close |%s| after output of the Bitmap\n",filename);
fflush(stdout);
  } /* end if filename != none */

} /* end GenerateBitmap() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void GenerateBitmapF(filename)
  char *filename;
{
 int i,j,k,ipixel;
 int idisplay,jdisplay;
 int PointerL;
 int Nonzero;
 int Fillin;

 FILE *fd;

  ntot=0;

/* initialize the display bitmap to all white */
  for (i=1;i<=ndisplay;i++)
   for (j=1;j<=ndisplay;j++)
    bitmap[j][i] = ZERO;

/* enter the diagonal */
 for (k=1; k<=n; k++) /* for the diagonal */
  {
   idisplay = ((k-1)/ncomb)+1;
   jdisplay = ((k-1)/ncomb)+1;
   ntot++;
   bitmap[idisplay][jdisplay]=NONZERO;
  } /* end for k */

/* calculate the display bitmap */
 for (k=1; k<=nadjlistF; k++) /* for all edges (including fillin) */
  {
   i = adjlistF[k].l.NodeNum;
   j = adjlistF[k].r.NodeNum;
   Fillin  = adjlistF[k].fillin;
   idisplay = ((i-1)/ncomb)+1;
   jdisplay = ((j-1)/ncomb)+1;
   if (Fillin == NO)
    {
     ntot += 2;
     if (NONZERO < bitmap[jdisplay][idisplay])
      {
       bitmap[jdisplay][idisplay]=NONZERO;
       bitmap[idisplay][jdisplay]=NONZERO;
      } /* end if */
    } /* end if */
   else
    {
     if (FILLIN < bitmap[jdisplay][idisplay])
      {
       bitmap[jdisplay][idisplay]=FILLIN;
       bitmap[idisplay][jdisplay]=FILLIN;
      } /* end if */
    } /* end else */
  } /* end for k */

printf("\nIn GenerateBitmap() - filename = |%s| \n",filename);
fflush(stdout);

  if (strcmp(filename,"none") != 0)
   {
    fd = fopen(filename, "w");

printf("Open |%s| for output of the Bitmap\n",filename);
fflush(stdout);

    for (i=1;i<=ndisplay;i++)
     for (j=1;j<=ndisplay;j++)
      {
       ipixel = bitmap[j][i];
       if( (ipixel < 0) || (ipixel > 255) )
        {
printf("error in writting bitmap[%d][%d]=|%d| to %s\n",j,i,ipixel);
fflush(stdout);
         ipixel = 0;
        } /* end if */
       fputc(ipixel,fd);
      }

    fclose(fd);

printf("Close |%s| after output of the Bitmap\n",filename);
fflush(stdout);
  } /* end if filename != none */

} /* end GenerateBitmapF() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PrintL(title)
 char *title;
{
 int i,j;
 int PointerL;
 int Nonzero;
 int Fillin;

 printf("\nSummary statistics for matrix fillin\n");
 printf("\tNumber of nonzeros in L                                    = %d \n",(n+nadjlist));
 printf("\tNumber of places in the entire matrix                      = %d \n",(n*n));
 printf("\tNumber of nonzeros in the strictly lower triangular matrix = %d \n",(nadjlist));
 printf("\tNumber of places in the strictly lower triangular matrix   = %d \n",((n-1)*n/2));
 printf("\tPercentage of non-zeros in the matrix                      = %f \n",
               ((double) (nadjlist * 100)/ (double) ((n-1)*n/2)));
 printf("\tNumber of fillins                                          = %d \n",0);
 fflush(stdout);

 if (XDISPLAY == YES)
   {
    printf("\n\n start X-Window display %s\n\n",title);
    Ximage(ndisplay,ndisplay,title);
    printf("\n\n end X-Window display %s\n\n",title);
    fflush(stdout);
   } /* end if */

} /* end PrintL() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PrintLF(title)
 char *title;
{
 int i,j;
 int PointerL;
 int Nonzero;
 int Fillin;

 printf("\nSummary statistics for matrix fillin\n");
 printf("\tNumber of nonzeros in L                                    = %d \n",(n+nadjlistF));
 printf("\tNumber of places in the entire matrix                      = %d \n",(n*n));
 printf("\tNumber of nonzeros in the strictly lower triangular matrix = %d \n",(nadjlistF));
 printf("\tNumber of places in the strictly lower triangular matrix   = %d \n",((n-1)*n/2));
 printf("\tPercentage of non-zeros in the matrix                      = %f \n",
               ((double) ((nadjlistF) * 100)/ (double) ((n-1)*n/2)));
 printf("\tNumber of fillins                                          = %d \n",FillinCount);
 fflush(stdout);

 if (XDISPLAY == YES)
   {
    printf("\n\n start X-Window display %s\n\n",title);
    Ximage(ndisplay,ndisplay,title);
    printf("\n\n end X-Window display %s\n\n",title);
    fflush(stdout);
   } /* end if */

} /* end PrintLF() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void FillinBreakdown()
{
  int i,j,k,blockstart,blockend;
  int borderstart,borderend;
  int fillin,fillinblocks,fillinborder;
  SoSElement *soselement;
  SetElement *element;
  Set *N1i;

/* initialize the values of block/border edges */
  blockstart  = 1;
  blockend    = (n - N2->n);
  borderstart = (n - N2->n) + 1;
  borderend   = n;
  fillin       = 0;
  fillinblocks = 0;
  fillinborder = 0;

 for (k=1; k<=nadjlistF; k++) /* for all non-zero elements (including fillin) */
   {
    if (adjlistF[k].fillin == YES)
     {
/* examine the blocks and determine the number of fillins */
    if ( (adjlistF[k].l.NodeNum < borderstart) && (adjlistF[k].r.NodeNum < borderstart) )
      fillinblocks++;
/* calculate the number of fillin in the last block */
    else if ( (adjlistF[k].l.NodeNum >= borderstart) && (adjlistF[k].r.NodeNum >= borderstart) )
           fillin++;
/* calculate the number of fillin in the borders */
    else
      fillinborder++;
    } /* end if adjlistF[k].fillin == YES */
   } /* end for k */

  printf("\nTotal fillin for blocks \t= %d \n",fillinblocks);
  printf("Total fillin for border \t= %d \n",fillinborder);
  printf("Fillin in the last block \t= %d \n",fillin);
  printf("Total fillin in the strictly lower diagonal matrix = %d \n",
          (fillinblocks+fillinborder+fillin) );

/* check the total amount of fillin */
  fillin = 0;
 for (k=1; k<=nadjlistF; k++) /* for all non-zero elements (including fillin) */
   {
    if (adjlistF[k].fillin == YES) fillin++;
   } /* end for k */

  printf("Total fillin in the strictly lower diagonal matrix = %d \n\n",fillin);
  fflush(stdout);

} /* end FillinBreakdown() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void Factor()
{
  int i,j,k,jk,ji,nOpsTotal;
  SetElement *elementColk,*elementRowk,*elementRowi;

/* initialize values */
  nOpsTotal = 0;

/* calculate the number of operations */
  for (k=1; k<=n; k++) /* along diagonal */
   {
    elementColk = SparseMatrix[k].Col->First;
    while (elementColk != NULL)
     {
      i = elementColk->Element;
      elementRowk = SparseMatrix[k].Row->First;
      while (elementRowk != NULL)
       {
        jk = elementRowk->Element;
        if (jk >= k)
	  {
           printf("--------- jk=%d - k=%d \n",jk,k);
           fflush(stdout);
	  } /* end if */
        elementRowi = SparseMatrix[i].Row->First;
        while (elementRowi != NULL)
	 {
          ji = elementRowi->Element;
/*
          if ( (ji == jk) && (ji < k) )
*/
          if (ji == jk)

	   {
            nOpsTotal++;
            break;
           } /* end if */
          elementRowi = elementRowi->Next;
	 } /* end while elementRowi != NULL*/
        elementRowk = elementRowk->Next;
       } /* end while elementRowk != NULL */
      elementColk = elementColk->Next;
     } /* end while elementColk != NULL */
   } /* end for k */

printf("\nFactor - nOpsTotal=%d \n\n",nOpsTotal);
fflush(stdout);

} /* end Factor() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void FactorBBDM()
{
  int i,j,k,jk,ji,ik,blockstart,blockend;
  int nOpsTotalFac,nOpsIndepFac,nOpsBlockFac,nOpsBorderFac;
  int nOpsTotalFrBs,nOpsBlockFrBs,nOpsBorderFrBs;
  int borderstart,borderend;
  SoSElement *soselement;
  SetElement *elementColk,*elementRowk,*elementRowi;
  Set *N1i;

/* initialize the values of block/border edges */
  blockstart  = blockend = 0;
  borderstart = (n - N2->n) + 1;
  borderend   = n;
  nOpsLastBlockFac = nOpsLastBlockFrBs = 0;
  nOpsTotalFac  = nOpsBlockFac  = nOpsBorderFac  = nOpsIndepFac  = 0;
  nOpsTotalFrBs = nOpsBlockFrBs = nOpsBorderFrBs = 0;

/* allocate memory for the vector of independent blocks */
  i = (n - N2->n);
  IndBlk = (IndBlocks *) calloc((i+1),sizeof(IndBlocks));
  if (IndBlk == NULL)
   {
    printf("can't allocate memory for IndBlk[] \n");
    exit(1);
   } /* end if */

/* set up the vector IndBlk[] */
  j = 0;
  soselement = N1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    blockstart = blockend + 1;
    blockend += N1i->n;
    if ( N1i->n <= 0)
      {
       printf("For cluster %s - N1i->n=%d \n",N1i->Name,N1i->n);
       PrintSet(N1i);
       fflush(stdout);
      } /* end if */
/*
printf("for cluster %s  --  blockstart=%d  --  blockend=%d \n",N1i->Name,blockstart,blockend);
fflush(stdout);
*/
    for (k=blockstart; k<=blockend; k++) /* along diagonal */
     {
      j++;
      IndBlk[j].N1i = N1i;
      IndBlk[j].blockstart = blockstart;
      IndBlk[j].blockend   = blockend;
     } /* end for k */

    soselement = soselement->Next;
   } /* end while soselement != NULL */

  if (i != j)
    {
     printf("(i & j should be equal!) j=%d  --  i=%d \n",j,i);
     fflush(stdout);
    } /* end if */

/*        k  is the index for the diagonal
          i  is an element of column k
	  jk is an element of row k
	  ji is an element of row i
*/
/* calculate the number of operations in Forward reduction 
   and Backward substitution and for Gauss-Seidel */
  for (k=1; k<=n; k++) /* along diagonal */
  {
/*
printf("Calc ops in FrBs --- k=%d -  \n",k);
fflush(stdout);
*/
/* these data structures are for the rows in the strictly 
   lower diagonal portion of the matrix */
   elementRowk = SparseMatrix[k].Row->First;
   while (elementRowk != NULL)
     {
      jk = elementRowk->Element;
/*
printf("\t\tjk=%d -  \n",jk);
fflush(stdout);
*/
/* calculate the number of multiply/adds in this row of a block */
      if (k < borderstart)
	{
         (IndBlk[k].N1i->nOpsBlockFrBs)++;
         (IndBlk[k].N1i->nOpsTotalFrBs)++;
         nOpsBlockFrBs++;
         nOpsTotalFrBs++;
	} /* end if k is in independent blocks */
      else
/* calculate the number of multiply/adds in this row of a border */
	{
         if (jk < borderstart)
	   {
            (IndBlk[jk].N1i->nOpsBorderFrBs)++;
            (IndBlk[jk].N1i->nOpsTotalFrBs)++;
            nOpsBorderFrBs++;
            nOpsTotalFrBs++;
	   } /* end if jk < borderstart */
         else
/* calculate the number of multiply/adds in this row of the last block */
	   {
            nOpsLastBlockFrBs++;
	   } /* end if jk >= borderstart */
	} /* end if k is in the border */
      elementRowk = elementRowk->Next;
     } /* end while elementRowk != NULL */
/* these data structures are for the rows in the strictly 
   upper diagonal portion of the matrix - the matrix is
   symmetric so the columns in the strictly lower diagonal
   portion of the matrix are the same as the rows in the
   strictly upper portion of the matrix */
   elementColk = SparseMatrix[k].Col->First;
   while (elementColk != NULL)
     {
      ik = elementColk->Element;
/* calculate the number of multiply/adds in this row of a block */
      if (k < borderstart)
	{
         (IndBlk[k].N1i->nOpsBlockFrBs)++;
         (IndBlk[k].N1i->nOpsTotalFrBs)++;
         nOpsBlockFrBs++;
         nOpsTotalFrBs++;
	} /* end if k is in independent blocks */
      else
/* calculate the number of multiply/adds in this row of the last block */
	{
         nOpsLastBlockFrBs++;
	} /* end if k is in the last block */
      elementColk = elementColk->Next;
     } /* end while elementRowk != NULL */
  } /* end for k */

/* calculate the number of operations in the factorization */
  for (k=1; k<=n; k++) /* along diagonal */
  {
    elementColk = SparseMatrix[k].Col->First;
    while (elementColk != NULL)
     {
      i = elementColk->Element;
      elementRowk = SparseMatrix[k].Row->First;
      while (elementRowk != NULL)
       {
        jk = elementRowk->Element;
        elementRowi = SparseMatrix[i].Row->First;
        while (elementRowi != NULL)
	 {
          ji = elementRowi->Element;
/*
          if ( (ji == jk) && (ji < k) )
*/
          if (ji == jk) 
	   {
            nOpsTotalFac++;
/*----------------------------------------------------------------------------*/
/* calculate the number of operations in the diagonal block and the border */
            if (k < borderstart)
             {
              (IndBlk[k].N1i->nOpsBlockFac)++;
              (IndBlk[k].N1i->nOpsTotalFac)++;
              nOpsIndepFac++;
              nOpsBlockFac++;
              if ( (ji < IndBlk[k].blockstart) || (ji > IndBlk[k].blockend) )
		{
                 printf("(ji not in a block!) - k=%d - i=%d - ji=%d - blockstart=%d - blockend=%d \n",
                      k,i,ji,IndBlk[i].blockstart,IndBlk[i].blockend);
                 fflush(stdout);
		} /* end if */
              break;
             } /* end if (k < borderstart) */
/* factor the borders before the last block */
            else if ( (k >= borderstart) && (ji < borderstart) )
             {
              (IndBlk[ji].N1i->nOpsBorderFac)++;
              (IndBlk[ji].N1i->nOpsTotalFac)++;
              nOpsIndepFac++;
              nOpsBorderFac++;
              break;
             } /* end else if (k >= borderstart) && (ji < borderstart) */
/* factor the last block */
            else if ( (k >= borderstart) && (ji >= borderstart) )
             {
              nOpsLastBlockFac++;
              break;
             } /* end else if (k >= borderstart) && (ji >= borderstart) */
            else
	      {
               printf("(value not accounted for!) - k=%d - i=%d - ji=%d - blockstart=%d - blockend=%d \n",
                      k,i,ji,IndBlk[i].blockstart,IndBlk[i].blockend);
                 fflush(stdout);
              } /* end else */
/*----------------------------------------------------------------------------*/
           } /* end if (ji == jk) && (ji < k) */
         elementRowi = elementRowi->Next;
	} /* end while elementRowi != NULL*/
       elementRowk = elementRowk->Next;
      } /* end while elementRowk != NULL */
     elementColk = elementColk->Next;
    } /* end while elementColk != NULL */
  } /* end for k */

  soselement = N1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    N1i->nOpsTotal = ((alpha*((double) N1i->nOpsTotalFac)) + 
                     ((1.0-alpha)*((double) N1i->nOpsTotalFrBs)));
    soselement = soselement->Next;
   } /* end while soselement !=NULL */

printf("\nOperations Counts for Factorization \n");
printf("\t(LastBlock - nOpsLastBlockFac=%d \n",nOpsLastBlockFac);
printf("\tFillinCount = %d \n",FillinCount);

printf("\tFactor - nOpsIndepFac=%d \n",nOpsIndepFac);
printf("\tFactor - nOpsBlockFac=%d \n",nOpsBlockFac);
printf("\tFactor - nOpsBorderFac=%d \n",nOpsBorderFac);
printf("\tFactor - nOpsTotalFac=%d \n",nOpsTotalFac);

printf("\nOperations Counts for Forward reduction/Backward Substitution \n");
printf("\tFrBs - nOpsBlockFrBs=%d \n",nOpsBlockFrBs);
printf("\tFrBs - nOpsBorderFrBs=%d \n",nOpsBorderFrBs);
printf("\tFrBs - nOpsTotalFrBs=%d \n",nOpsTotalFrBs);

printf("\n\t(LastBlock - nOpsLastBlockFrBs=%d \n",nOpsLastBlockFrBs);
printf("\t(LastBlock - nOpsLastBlockFac=%d \n",nOpsLastBlockFac);

printf("\nTotal operations Count=%lf --- alpha=%lf \n",
          ((alpha*((double) nOpsTotalFac)) + ((1.0-alpha)*((double) nOpsTotalFrBs))),alpha);
fflush(stdout);

  free(IndBlk);

} /* end FactorBBDM() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PigeonHoleLoadBalance(tN1,FileNameA,FileNameb,FileNameN2,namekey,nprocs)
  SoSets *tN1;
  char *FileNameA,*FileNameb,*FileNameN2,*namekey;
  int nprocs;
{
  int i,j,nclusters,MinOpsLoc,nlb;
  int nOpsTotalFac,nOpsBlockFac,nOpsBorderFac;
  int nOpsTotalFrBs,nOpsBlockFrBs,nOpsBorderFrBs;
  double nOpsTotal,MaxOps,MinOps;
  double totindep,minindep,maxindep,totblocks,totborder;
  int totindepFac,minindepFac,maxindepFac,totblockFac,totborderFac;
  int totindepFrBs,minindepFrBs,maxindepFrBs,totblockFrBs,totborderFrBs;
  char filenameA[40],filenameb[40];
  char filename[40];
  char *string = "SuperCluster";
  char *name;
  SoSElement *soselement;
  SetElement *element;
  Set *tN1i,*MaxOpsSet;
  SoSets *SC;

printf("PigeonHoleLoadBalance --- namekey=%s - nprocs=%d \n",namekey,nprocs);
fflush(stdout);

  if (loadbal != NULL) free(loadbal);
  loadbal = (LoadBalance *) calloc((nprocs+1),sizeof(LoadBalance));
  if (loadbal == NULL)
   {
    printf("can't allocate memory for loadbal[] \n");
    exit(1);
   } /* end if */

/* reset the SuperCluster values to NO */
/*
PrintSetOfSets(tN1);

printf("reset the SuperCluster values to NO \n");
fflush(stdout);
*/
  nOpsTotalFac  = nOpsBlockFac  = nOpsBorderFac  = 0;
  nOpsTotalFrBs = nOpsBlockFrBs = nOpsBorderFrBs = 0;
  soselement = tN1->First;
  while (soselement != NULL)
   {
    tN1i = soselement->set;
    tN1i->SuperCluster = NO;
    nOpsTotalFac   += tN1i->nOpsTotalFac;
    nOpsBlockFac   += tN1i->nOpsBlockFac;
    nOpsBorderFac  += tN1i->nOpsBorderFac;
    nOpsTotalFrBs  += tN1i->nOpsTotalFrBs;
    nOpsBlockFrBs  += tN1i->nOpsBlockFrBs;
    nOpsBorderFrBs += tN1i->nOpsBorderFrBs;
    nOpsTotal      += tN1i->nOpsTotal;
    soselement = soselement->Next;
   } /* end while soselement !=NULL */

printf("nOpsTotalFac=%d - nOpsBlockFac=%d - nOpsBorderFac=%d \n",
        nOpsTotalFac,nOpsBlockFac,nOpsBorderFac);
printf("nOpsTotalFrBs=%d - nOpsBlockFrBs=%d - nOpsBorderFrBs=%d \n",
        nOpsTotalFrBs,nOpsBlockFrBs,nOpsBorderFrBs);
printf("  alpha=%lf - nOpsTotal=%lf \n",alpha,nOpsTotal);
fflush(stdout);

/* load the first number of clusters into the pigeon holes */

printf("load the first number of clusters into the pigeon holes\n");
fflush(stdout);

  nclusters = tN1->n;
  if (nclusters < nprocs)
   {
    printf(" The number of clusters is less than the number of pigeon holes\n");
    printf("  number of clusters = %d  --  number of processors = %d \n",nclusters,nprocs);
    exit(1);
   } /* end if number of sets < nprocs */
  else
   {
    for (i=1; i<=nprocs; i++)
     {
      MaxOps = 0.0;
      soselement = tN1->First;
      while (soselement != NULL)
       {
        tN1i = soselement->set;
/*
printf("%s --- tN1i->nOpsTotal = %lf --- ",tN1i->Name,tN1i->nOpsTotal);
fflush(stdout);
printf(" --- tN1i->SuperCluster = %d\n",tN1i->SuperCluster);
fflush(stdout);
*/
        if ( (tN1i->nOpsTotal > MaxOps) && (tN1i->SuperCluster == NO) )
         {
          MaxOps = tN1i->nOpsTotal;
          MaxOpsSet = tN1i;
/*
printf("MaxOps = %lf -- ",MaxOps);
fflush(stdout);
printf(" -- MaxOpsSet = %s\n",tN1i->Name);
fflush(stdout);
*/
         } /* end if */
        soselement = soselement->Next;
       } /* end while soselement !=NULL */
      name = (char *) calloc(16,sizeof(char));
      if (name == NULL)
       {
        printf("can't allocate memory for name[] \n");
        exit(1);
       } /* end if */
      for (j=0; j<12; j++)
        name[j] = string[j];
      name[12] = '-';
      name[13] = (char) (((i%100)/10) + 48);
      name[14] = (char) ((i%10) + 48);
      name[15] = '\0';
      SC = NULL;
      SC = InitSetOfSets(SC,name);
      AddSetToSetOfSets(SC,MaxOpsSet);
      loadbal[i].SuperCluster = SC;
/*
PrintSetOfSets(loadbal[i].SuperCluster);
*/
      loadbal[i].nOpsBlockFac  = MaxOpsSet->nOpsBlockFac;
      loadbal[i].nOpsBorderFac = MaxOpsSet->nOpsBorderFac;
      loadbal[i].nOpsTotalFac  = MaxOpsSet->nOpsTotalFac;
      loadbal[i].nOpsBlockFrBs  = MaxOpsSet->nOpsBlockFrBs;
      loadbal[i].nOpsBorderFrBs = MaxOpsSet->nOpsBorderFrBs;
      loadbal[i].nOpsTotalFrBs  = MaxOpsSet->nOpsTotalFrBs;
      loadbal[i].nOpsTotal  = MaxOpsSet->nOpsTotal;
      MaxOpsSet->SuperCluster = i;
     } /* end for i */
   } /* end if number of sets >= nprocs */

/* place the remaining clusters into super clusters in a manner to
   balance the load - place the largest remaining cluster into the
   smallest super cluster - largest remaining cluster has the greatest 
   nOpsBlock value and smallest super block has the least total nOpsBlock */

printf("place the remaining clusters into super clusters\n");
fflush(stdout);

  for (i=1; i<=(nclusters-nprocs); i++)
   {
/*
printf("find the largest remaining cluster --- i=%d  \n",i);
fflush(stdout);
*/
/* find the largest remaining cluster */
    MaxOps = -1.0;
    MaxOpsSet = NULL;
    soselement = tN1->First;
    while (soselement != NULL)
     {
      tN1i = soselement->set;
      if ( (tN1i->nOpsTotal > MaxOps) && (tN1i->SuperCluster == NO) )
       {
        MaxOps = tN1i->nOpsTotal;
        MaxOpsSet = tN1i;
       } /* end if */
      soselement = soselement->Next;
     } /* end while soselement !=NULL */
/*
printf("\tfind the smallest super cluster \n");
fflush(stdout);
*/
/*
PrintSet(MaxOpsSet);
*/
/* find the smallest super cluster */
    MinOps = (double) INFI;
    MinOpsLoc = 0;
    for (j=1; j<=nprocs; j++)
     {
      if (loadbal[j].nOpsTotal < MinOps)
       {
        MinOps = loadbal[j].nOpsTotal;
        MinOpsLoc = j;
       } /* end if */
     } /* end for j */
/*
printf("\tupdate the values \n");
fflush(stdout);
*/
/* update the values */
    AddSetToSetOfSets(loadbal[MinOpsLoc].SuperCluster,MaxOpsSet);
    loadbal[MinOpsLoc].nOpsBlockFac   += MaxOpsSet->nOpsBlockFac;
    loadbal[MinOpsLoc].nOpsBorderFac  += MaxOpsSet->nOpsBorderFac;
    loadbal[MinOpsLoc].nOpsTotalFac   += MaxOpsSet->nOpsTotalFac;
    loadbal[MinOpsLoc].nOpsBlockFrBs  += MaxOpsSet->nOpsBlockFrBs;
    loadbal[MinOpsLoc].nOpsBorderFrBs += MaxOpsSet->nOpsBorderFrBs;
    loadbal[MinOpsLoc].nOpsTotalFrBs  += MaxOpsSet->nOpsTotalFrBs;
    loadbal[MinOpsLoc].nOpsTotal      += MaxOpsSet->nOpsTotal;
    MaxOpsSet->SuperCluster = MinOpsLoc;
   } /* end for i */
/*
printf("\tprint out the super clusters \n");
fflush(stdout);
*/
/* print out the super clusters */
  totindepFac  = totblockFac  = totborderFac  = 0;
  minindepFac = INFI;
  maxindepFac = 0;
  totindepFrBs = totblockFrBs = totborderFrBs = 0;
  minindepFrBs = INFI;
  maxindepFrBs = 0;
  totindep = 0.0;
  minindep = (double) INFI;
  maxindep = 0.0;
  for (i=1; i<=nprocs; i++)
   {

    printf("\nSuper Cluster Number %d --- for nprocs=%d - alpha=%lf\n",i,nprocs,alpha);
    printf("\tNumber of operations (Fac) = %d \n",loadbal[i].nOpsTotalFac);
    printf("\tNumber of operations (FrBs = %d \n",loadbal[i].nOpsTotalFrBs);
    printf("\tTotal number of operations = %lf \n",loadbal[i].nOpsTotal);
    fflush(stdout);

/*
    PrintSetOfSets(loadbal[i].SuperCluster);
*/
    totindepFac  += loadbal[i].nOpsTotalFac;
    totblockFac  += loadbal[i].nOpsBlockFac;
    totborderFac += loadbal[i].nOpsBorderFac;
    if (loadbal[i].nOpsTotalFac < minindepFac) minindepFac = loadbal[i].nOpsTotalFac;
    if (loadbal[i].nOpsTotalFac > maxindepFac) maxindepFac = loadbal[i].nOpsTotalFac;
    totindepFrBs  += loadbal[i].nOpsTotalFrBs;
    totblockFrBs  += loadbal[i].nOpsBlockFrBs;
    totborderFrBs += loadbal[i].nOpsBorderFrBs;
    if (loadbal[i].nOpsTotalFrBs < minindepFrBs) minindepFrBs = loadbal[i].nOpsTotalFrBs;
    if (loadbal[i].nOpsTotalFrBs > maxindepFrBs) maxindepFrBs = loadbal[i].nOpsTotalFrBs;
    totindep  += loadbal[i].nOpsTotal;
    if (loadbal[i].nOpsTotal < minindep) minindep = loadbal[i].nOpsTotal;
    if (loadbal[i].nOpsTotal > maxindep) maxindep = loadbal[i].nOpsTotal;
   } /* end for i */

  nlb = N2->n;
  printf("\nSummary Statistics for nprocs=%d \n",nprocs);
  printf("\tTotal number of operations in the independent blocks = %lf \n",totindep);
  printf("\t\tTotal number of operations in the independent blocks (Fac) = %d \n",totindepFac);
  printf("\t\tTotal number of operations in the blocks            (Fac) = %d \n",totblockFac);
  printf("\t\tTotal number of operations in the borders           (Fac) = %d \n\n",totborderFac);
  printf("\t\tTotal number of operations in the independent blocks (FrBs) = %d \n",totindepFrBs);
  printf("\t\tTotal number of operations in the blocks             (FrBs) = %d \n",totblockFrBs);
  printf("\t\tTotal number of operations in the borders            (FrBs) = %d \n\n",totborderFrBs);
  printf("\tTotal number of operations in the lastblock (sparse frbs)     = %d \n",nOpsLastBlockFrBs);
  printf("\tTotal number of operations in the lastblock (sparse fac)      = %d \n",nOpsLastBlockFac);
  printf("\tTotal number of operations in the lastblock (dense)           = %d \n\n",(((nlb*nlb*nlb)-nlb)/3));
  printf("\tTotal number of operations (sparse)                  = %lf \n",(totindep + nOpsLastBlockFac));
  printf("\tTotal number of operations (dense)                   = %lf \n\n",(totindep + (((nlb*nlb*nlb)-nlb)/3)));
  printf("\tMinimum number of operations in a cluster            = %lf \n",minindep);
  printf("\tMaximum number of operations in a cluster            = %lf \n",maxindep);
  printf("\t(Maximum-Minimum) operations in a cluster            = %lf \n\n",(maxindep - minindep));
  printf("\tPercent of operations in the clusters (sparse)       = %f \n",
                     (double) ((double) (totindep * 100) / (double) (totindep + nOpsLastBlockFac)));
  printf("\tPercent of operations in the clusters (dense)        = %f \n",
                     (double) ((double) (totindep * 100) / (double) (totindep + (((nlb*nlb*nlb)-nlb)/3))));


/* place the superclusters in contigious locations */
/* initialize the set of sets sN1 */
  sN1 = NULL;
  sN1 = InitSetOfSets(sN1,"SuperClusters");
  if (proc != NULL) free(proc);
  proc = (Proc *) calloc((nprocs+1),sizeof(Proc));
  if (proc == NULL)
   {
    printf("can't allocate memory for proc[] \n");
    exit(1);
   } /* end if */

/* copy the sets to the set of set sN1 so the superclusters
   are in adjacent order */
   proc[0].BlockPtr       = 0;
   proc[0].DiagPtr        = 0;
   proc[0].BlockRowPtr    = 0;
   proc[0].LowerBorderPtr = 0;
   proc[0].BorderRowPtr   = 0;
  for (i=nprocs; i>=1; i--)
    {
     proc[i].BlockPtr       = 0;
     proc[i].DiagPtr        = 0;
     proc[i].BlockRowPtr    = 0;
     proc[i].LowerBorderPtr = 0;
     proc[i].BorderRowPtr   = 0;
     SC = loadbal[i].SuperCluster;

printf("(loadbal[%d].SuperCluster)->Name=%s - (loadbal[%d].SuperCluster)->n=%d \n",
         i,(loadbal[i].SuperCluster)->Name,i,(loadbal[i].SuperCluster)->n);
fflush(stdout);

     soselement = SC->First;
     while (soselement != NULL)
      {
       tN1i = soselement->set;
       AddSetToSetOfSets(sN1,tN1i);
       proc[i].BlockPtr       += 1;
       proc[i].DiagPtr        += tN1i->n;

printf("\ttN1i->Name=%s - n=%d - nOTFac=%d - nOTFrBs=%d - nOT=%lf \n",
               tN1i->Name,tN1i->n,tN1i->nOpsTotalFac,tN1i->nOpsTotalFrBs,tN1i->nOpsTotal);
fflush(stdout);

       soselement = soselement->Next;
      } /* soselement != NULL */
    } /* end for i */
/* sum values to make proc[].BlockPtr contigious */

printf("\tproc[0].BlockPtr=%d \n",proc[0].BlockPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].BlockPtr += proc[i-1].BlockPtr;

printf("\tproc[%d].BlockPtr=%d \n",i,proc[i].BlockPtr);
fflush(stdout);

    } /* end for i */
/* sum values to make proc[].DiagPtr contigious */

printf("\tproc[0].DiagPtr=%d \n",proc[0].DiagPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].DiagPtr += proc[i-1].DiagPtr;

printf("\tproc[%d].DiagPtr=%d \n",i,proc[i].DiagPtr);
fflush(stdout);

    } /* end for i */
/* reset the adjacency list */
  ResetAdjList();
/*
PrintSetOfSets(sN1);
*/
/* call ReorderNetworkClusters() to generate a new adjlistF */

printf("ReorderNetworkClusters() \n");
fflush(stdout);

  ReorderNetworkClusters(sN1);

printf("GenerateBitmapF \n");
fflush(stdout);
  GenerateBitmapF(MakeOutputFilename(namekey,filename,(100+nprocs))); /* produce a bitmap of the fillin pattern */

printf("PrintLF \n");
fflush(stdout);
  PrintLF("NiMo Data"); /* produce an X Window display of the fillin pattern */

/* generate data for bdb-form Choleski solver */

printf("BDBFormData() \n");
fflush(stdout);

  BDBFormData(nprocs);
/* write data for bdb-for Choleski solver to a file */

printf("OutputBDBFormData() \n");
fflush(stdout);

  OutputBDBFormData(MakeOutputFilename(FileNameA,filenameA,nprocs),
                    MakeOutputFilename(FileNameb,filenameb,nprocs),
                    FileNameN2,nprocs);

} /* end PigeonfactorHoleLoadBalance() */

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void BDBFormData(nprocs)
  int nprocs;
{
  int i,j,k,l,m,done;
  int nBD,nLB,nBor;
  int nlastblockstart,nlastblockend;
  Set *N1i;
  SoSElement *soselement;
  SetElement *element;

  nlastblockstart = n - (N2->n) + 1;
  nlastblockend   = n;

printf("nlastblockstart=%d - nlastblockend=%d \n",nlastblockstart,nlastblockend);
fflush(stdout);

/* determine the size of the data-structure elements */
/* allocate memory for the bdb-form sparse matrix data structures */
  nblocks = sN1->n;

printf("nblocks=%d  \n",nblocks);
fflush(stdout);

  if (blocks != NULL) free(blocks);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/
printf("determine blocks --- \n");
fflush(stdout);

  blocks[0].DiagPtr = 0;
  i = 1;
  soselement = sN1->First;
  while (soselement != NULL)
    {
     blocks[i].BlockID = i;
     blocks[i].LBSet   = NULL;
     blocks[i].LBSet   = InitSet(blocks[i].LBSet,"LastBlockSet");
     N1i = soselement->set;
     blocks[i].BlockSz = N1i->n;
     blocks[i].DiagPtr = blocks[i-1].DiagPtr + N1i->n;
     for (j=1;j<=nprocs;j++)
       {
        if ( (blocks[i-1].DiagPtr >= proc[j-1].DiagPtr) && 
             (blocks[i].DiagPtr   <= proc[j].DiagPtr) )
	  {
           blocks[i].ProcNum = j;
           break;
	  } /* end if */
       } /* end for l*/
     soselement = soselement->Next;
     i++;
    } /* soselement != NULL */

printf("BDBFormData() --- i=%d - nblocks=%d \n",i,nblocks);
fflush(stdout);

/*---------------------------------------------------*/
printf("determine diag --- \n");
fflush(stdout);

  ndiag = n;

printf("\t\tndiag=%d  \n",ndiag);
fflush(stdout);

  if (diag != NULL) free(diag);
  diag = (Diag *) calloc((ndiag+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1;i<=ndiag;i++)
    {
     diag[i].Value = ndiag + drand48();
     diag[i].BlockRowSet  = NULL;
     diag[i].BlockRowSet  = InitSet(diag[i].BlockRowSet,"BlockRowSet");
     diag[i].NextLxkSet   = NULL;
     diag[i].NextLxkSet   = InitSet(diag[i].NextLxkSet,"NextLxkSet");
     diag[i].NextLBxkSet  = NULL;
     diag[i].NextLBxkSet  = InitSet(diag[i].NextLBxkSet,"NextLBxkSet");
    } /* end for i */
/*---------------------------------------------------*/

printf("determine elements in BlockRowSet and nextLxkSet --- \n");
fflush(stdout);

 for (k=1; k<=nadjlistF; k++) /* for all non-zero elements (including fillin) */
   {
    if ( ( (adjlistF[k].l.NodeNum <  nlastblockstart) && (adjlistF[k].r.NodeNum <  nlastblockstart) ) ||
         ( (adjlistF[k].l.NodeNum >= nlastblockstart) && (adjlistF[k].r.NodeNum >= nlastblockstart) ) )
      {
       if (adjlistF[k].l.NodeNum < adjlistF[k].r.NodeNum)
         {
	  i = adjlistF[k].r.NodeNum;
          j = adjlistF[k].l.NodeNum;
	 } /* end if */
       else
	 {
	  i = adjlistF[k].l.NodeNum;
          j = adjlistF[k].r.NodeNum;
	 } /* end else */
	AddElementToSet(diag[i].BlockRowSet,j);
        if ( (adjlistF[k].l.NodeNum <  nlastblockstart) && (adjlistF[k].r.NodeNum <  nlastblockstart) )
	   AddElementToSet(diag[j].NextLxkSet, i);
      } /* end if in diagonal-bordered blocks or last block */
    else
      {
       if (adjlistF[k].l.NodeNum < adjlistF[k].r.NodeNum)
         {
	  i = adjlistF[k].r.NodeNum;
          j = adjlistF[k].l.NodeNum;
	 } /* end if */
       else
	 {
	  i = adjlistF[k].l.NodeNum;
          j = adjlistF[k].r.NodeNum;
	 } /* end else */
       AddElementToSet(diag[j].NextLBxkSet,i);
       for (l=1;l<=nblocks;l++)
	{
         if ( (j > blocks[l-1].DiagPtr) && (j <= blocks[l].DiagPtr) )
	   {
            AddElementToSet(blocks[l].LBSet,i);
            break;
	   } /* end if j is in a lower block */
	} /* end for l */
      } /* end if in lower border */
   } /* end for k */
/*---------------------------------------------------*/
 nBD = nLB = nBor = 0;
 for (k=1; k<=nadjlistF; k++) /* for all non-zero elements (including fillin) */
   {
    if ( (adjlistF[k].l.NodeNum <  nlastblockstart) && (adjlistF[k].r.NodeNum <  nlastblockstart) )
      {
        nBD++;
      } /* end if in diagonal-bordered blocks */
    else if ( (adjlistF[k].l.NodeNum >= nlastblockstart) && (adjlistF[k].r.NodeNum >= nlastblockstart) ) 
      {
        nLB++;
      } /* end if in last block */
    else
      {
        nBor++;
      } /* end if in lower border */
   } /* end for k */

printf("\tnBD=%d - nLB=%d - nBor=%d - \n",nBD,nLB,nBor);
fflush(stdout);

/*---------------------------------------------------*/

printf("determine blocks[].LBPtr --- \n");
fflush(stdout);

  blocks[0].LBPtr = 0;
  for (i=1;i<=nblocks;i++)
    {
     blocks[i].LBPtr = blocks[i-1].LBPtr + (blocks[i].LBSet)->n;
     for (j=1;j<=nprocs;j++)
       {
        if (( i > proc[j-1].BlockPtr) && (i <= proc[j].BlockPtr) )
	  {
           proc[j].LowerBorderPtr += (blocks[i].LBSet)->n;
           break;
	  } /* end if */
       } /* end for j */
    } /* end for i */
/* sum values to make proc[].LowerBorderPtr contigious */

printf("\tproc[0].LowerBorderPtr=%d \n",proc[0].LowerBorderPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].LowerBorderPtr += proc[i-1].LowerBorderPtr;

printf("\tproc[%d].LowerBorderPtr=%d \n",i,proc[i].LowerBorderPtr);
fflush(stdout);

    } /* end for i */
/*---------------------------------------------------*/

printf("determine diag[].PtrLix --- \n");
fflush(stdout);

  diag[0].PtrLix = 0;
  for (i=1;i<nlastblockstart;i++)
    {
     diag[i].PtrLix = diag[i-1].PtrLix + (diag[i].BlockRowSet)->n;
     for (j=1;j<=nprocs;j++)
       {
        if ( (i > proc[j-1].DiagPtr) && (i <= proc[j].DiagPtr) )
	  {
           proc[j].BlockRowPtr += (diag[i].BlockRowSet)->n;
           break;
	  } /* end if */
       } /* end for j */
    } /* end for i */
/* sum values to make proc[].BlockRowPtr contigious */

printf("\tproc[0].BlockRowPtr=%d \n",proc[0].BlockRowPtr);
fflush(stdout);

   for (i=1;i<=nprocs;i++)
    {
     proc[i].BlockRowPtr += proc[i-1].BlockRowPtr;

printf("\tproc[%d].BlockRowPtr=%d \n",i,proc[i].BlockRowPtr);
fflush(stdout);

    } /* end for i */
/*---------------------------------------------------*/

printf("determine blockrow --- \n");
fflush(stdout);

  nblockrow = 0;
  for (i=1;i<nlastblockstart;i++)
    nblockrow += (diag[i].BlockRowSet)->n;

printf("\t\t --- nblockrow=%d \n",nblockrow);
fflush(stdout);

  if (blockrow != NULL) free(blockrow);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  k = 1;
/*
printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);
*/
  for (i=1;i<nlastblockstart;i++)
    {
/*
printf("\t\t --- k=%d - i=%d - (diag[%d].BlockRowSet)->n=%d - nblockrow=%d \n",
          k,i,i,(diag[i].BlockRowSet)->n,nblockrow);
fflush(stdout);
*/
     while ((diag[i].BlockRowSet)->n > 0)
       {
        blockrow[k].Value = -1.0 * drand48();
        blockrow[k].i = i;
        blockrow[k].j = MinValueInSet(diag[i].BlockRowSet);
/*
printf("\t\t --- blockrow[%d].Value=%lf - blockrow[%d].i=%d - blockrow[%d].j=%d \n",
              k,blockrow[k].Value,k,blockrow[k].i,k,blockrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
    } /* end for i */

printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);
/*---------------------------------------------------*/

printf("determine lborder --- \n");
fflush(stdout);

  nlborder = 0;
  for (i=1;i<=nblocks;i++)
    nlborder += (blocks[i].LBSet)->n;

printf("\t\t --- nlborder=%d  \n",nlborder);
fflush(stdout);

  if (lborder != NULL) free(lborder);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */

printf("initialize BorderRowSet --- \n");
fflush(stdout);

  k = 1;
  for (i=1;i<=nblocks;i++)
    {
     while ((blocks[i].LBSet)->n > 0)
       {
        lborder[k].i = MinValueInSet(blocks[i].LBSet);
        lborder[k].BorderRowSet = NULL;
        lborder[k].BorderRowSet = InitSet(lborder[k].BorderRowSet,"BorderRowSet");
        k++;
       } /* end while */
    } /* end for i */
/*---------------------------------------------------*/

printf("determine elements of BorderRowSet --- \n");
fflush(stdout);
 
 for (k=1; k<=nadjlistF; k++) /* for all non-zero elements (including fillin) */
   {
    done = NO;
    if ( ( (adjlistF[k].l.NodeNum < nlastblockstart) && (adjlistF[k].r.NodeNum >= nlastblockstart) ) ||
         ( (adjlistF[k].l.NodeNum >= nlastblockstart) && (adjlistF[k].r.NodeNum < nlastblockstart) ) )
      {
      if ( (adjlistF[k].l.NodeNum < nlastblockstart) && (adjlistF[k].r.NodeNum >= nlastblockstart) )
        {
         i = adjlistF[k].r.NodeNum;
         j = adjlistF[k].l.NodeNum;
        } /* end if in border */
      else if ( (adjlistF[k].l.NodeNum >= nlastblockstart) && (adjlistF[k].r.NodeNum < nlastblockstart) )
        {
         i = adjlistF[k].l.NodeNum;
         j = adjlistF[k].r.NodeNum;
        } /* end if in border */
       for(l=1;( (l<=nblocks) && (done == NO) );l++) /* find the block */
         {
          if ( (j > blocks[l-1].DiagPtr) && (j <= blocks[l].DiagPtr))
	    {
/* find the location in lborder[] */
             for (m=(blocks[l-1].LBPtr+1);( (m<=blocks[l].LBPtr) && (done == NO) );m++)
	      {
               if (lborder[m].i == i)
	         {
                  AddElementToSet(lborder[m].BorderRowSet,j);
                  done = YES;
	         } /* end if in lborder[m] */
	      }   /* end for m */
	    } /* end if in block l */
         } /* end for l */
      } /* end if */
   } /* end for k */

  nborderrow = 0;
  for (i=1;i<=nlborder;i++)
   nborderrow += (lborder[i].BorderRowSet)->n;

printf("\t\t --- nborderrow=%d  \n",nborderrow);
fflush(stdout);

  if (borderrow != NULL) free(borderrow);
  borderrow = (BorderRow *) calloc((nborderrow+1),sizeof(BorderRow));
  if (borderrow == NULL)
   {
    printf("can't allocate memory for borderrow[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/

printf("determine values of borderrow --- \n");
fflush(stdout);

  for (i=1;i<=nlborder;i++)
   {
     for (j=1;j<=nprocs;j++)
       {
        if (( i > proc[j-1].LowerBorderPtr) && (i <= proc[j].LowerBorderPtr) )
	  {
           proc[j].BorderRowPtr += (lborder[i].BorderRowSet)->n;
           break;
	  } /* end if */
       } /* end for j */
   } /* end for i */
 /* sum values to make proc[].BorderRowPtr contigious */

printf("\tproc[0].BorderRowPtr=%d \n",proc[0].BorderRowPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].BorderRowPtr += proc[i-1].BorderRowPtr;

printf("\tproc[%d].BorderRowPtr=%d \n",i,proc[i].BorderRowPtr);
fflush(stdout);

    } /* end for i */
 
  lborder[0].PtrLBix = 0;
  k = 1;
  for (i=1;i<=nlborder;i++)
   {
    lborder[i].PtrLBix = lborder[i-1].PtrLBix + (lborder[i].BorderRowSet)->n;

/*
printf("\t\tk=%d - i=%d - lborder[%d].i=%d - lborder[%d].PtrLBix=%d - (lborder[%d].BorderRowSet)->n=%d \n",
             k,i,i,lborder[i].i,i,lborder[i].PtrLBix,i,(lborder[i].BorderRowSet)->n);
fflush(stdout);
*/
    while ((lborder[i].BorderRowSet)->n > 0)
      {
        borderrow[k].Value = -1.0 * drand48();
        borderrow[k].i = lborder[i].i;
        borderrow[k].j = MinValueInSet(lborder[i].BorderRowSet);
/*
printf("\t\ti=%d - k=%d - lborder[%d].i=%d - borderrow[%d].Value=%lf - borderrow[%d].i=%d - borderrow[%d].j=%d \n",
             i,k,i,lborder[i].i,k,borderrow[k].Value,k,borderrow[k].i,k,borderrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
   } /* end for i */
/*---------------------------------------------------*/

printf("determine diag[].NextLxk  --- \n");
fflush(stdout);

  for (k=1;k<nlastblockstart;k++)
   {
    if ((diag[k].NextLxkSet)->n > 0)
     {
      i = MinValueInSet(diag[k].NextLxkSet);
      for (j=(diag[i-1].PtrLix + 1);j<=diag[i].PtrLix;j++)
       {
        if ( (i == blockrow[j].i) && (k == blockrow[j].j) )
	  {
           diag[k].NextLxk = j;
           break;
	  } /* end if */
	} /* end for j */
      } /* end if */
    else
      {
       diag[k].NextLxk = 0;
      } /* end else */
   } /* end for k */
/*---------------------------------------------------*/

printf("determine diag[].NextLBxkSet --- \n");
fflush(stdout);

  for (k=1;k<nlastblockstart;k++)
   {
    if ((diag[k].NextLBxkSet)->n >0)
     {
      i = MinValueInSet(diag[k].NextLBxkSet);
      for (j=1;j<=nborderrow;j++)
       {
        if ( (i == borderrow[j].i) && (k == borderrow[j].j) )
	  {
           diag[k].NextLBxk = j;
           break;
	  } /* end if */
       } /* end for j */
     } /* end if */
    else
     {
      diag[k].NextLBxk = 0;
     } /* end else */
   } /* end for k */
/*---------------------------------------------------*/

printf("determine blockrow[k].NextLxk --- \n");
fflush(stdout);

  for (k=1;k<nblockrow;k++)
    {
      if ((diag[ blockrow[k].j ].NextLxkSet)->n > 0)
       {
        i = MinValueInSet(diag[ blockrow[k].j ].NextLxkSet);
        for (j=(diag[i-1].PtrLix + 1);j<=diag[i].PtrLix;j++)
         {
          if ( (i == blockrow[j].i) && (blockrow[k].j == blockrow[j].j) )
	   {
            blockrow[k].NextLxk = j;
            break;
	   } /* end if */
	 } /* end for j */
/*
printf("\t\t --- blockrow[%d].NextLxk=%d \n",k,blockrow[k].NextLxk);
fflush(stdout);
*/
       } /* end if */
      else
       {
        blockrow[k].NextLxk = 0;
/*
printf("\t\t --- blockrow[%d].NextLxk=%d - k=%d \n",k,blockrow[k].NextLxk,k);
fflush(stdout);
*/
       } /* end else */
    } /* end for k */
/*---------------------------------------------------*/

printf("determine values in borderrow[].NextLBxk --- \n");
fflush(stdout);
 
  for (k=1;k<=nborderrow;k++)
    {
     if ( (diag[ borderrow[k].j ].NextLBxkSet)->n > 0)
       {
        i = MinValueInSet(diag[ borderrow[k].j ].NextLBxkSet);
        for (j=(k+1);j<=nborderrow;j++)
	 {
          if ( (i == borderrow[j].i) && (borderrow[k].j == borderrow[j].j) )
	    {
             borderrow[k].NextLBxk = j;
             break;
	    } /* end if */
 	 } /* end for j */
       } /* end if */
     else
       {
        borderrow[k].NextLBxk = 0;
       } /* end else */
/*
printf("\t\tk=%d - borderrow[%d].NextLBxk=%d \n",
             k,k,borderrow[k].NextLBxk);
fflush(stdout);
*/
    } /* end for k */
/*---------------------------------------------------*/

printf("determine values in lastbk[]. --- \n");
fflush(stdout);
 
  ln = N2->n;
  lnnz = 0;
  for (i=nlastblockstart;i<=nlastblockend;i++)
    lnnz += (diag[i].BlockRowSet)->n;
/*
printf("\tln=%d - lnnz=%d \n",ln,lnnz);
fflush(stdout);
*/
  lnnz += ln;
  if (lastbk != NULL) free(lastbk);
  lastbk = (LastBlock *) calloc((lnnz+1),sizeof(LastBlock));
  if (lastbk == NULL)
   {
    printf("can't allocate memory for lastbk[] \n");
    exit(1);
   } /* end if */

printf("\t\tln=%d - lnnz=%d \n",ln,lnnz);
fflush(stdout);

  k = 1;
  for (i=nlastblockstart;i<=nlastblockend;i++)
   {
/*
printf("i=%d - ",i);
fflush(stdout);
*/
    lastbk[k].Value = ndiag + drand48();
    lastbk[k].i     = i;
    lastbk[k].j     = i;
    k++;
/*
printf("lastbk[%d].Value=%d  - lastbk[%d].i=%d - lastbk[%d].j=%d - (diag[i].BlockRowSet)->n=%d\n",
              k,lastbk[k].Value,k,lastbk[k].i,k,lastbk[k].j,i,(diag[i].BlockRowSet)->n);
fflush(stdout);
*/
     while ((diag[i].BlockRowSet)->n > 0)
       {
        lastbk[k].Value = -1.0 * drand48();
        lastbk[k].i     = i;
        lastbk[k].j     = MinValueInSet(diag[i].BlockRowSet);
        k++;
       } /* end while */
   } /* end for i */

printf("\t\tln=%d - lnnz=%d = k=%d \n",ln,lnnz,k);
fflush(stdout);

/* do a check on sets to see that all values are accounted for*/
  k = 0;
  for (i=1;i<=nblocks;i++)
    {
     k += (blocks[i].LBSet)->n;
    } /* end for i */
  printf("Check on set blocks[].LBSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=ndiag;i++)
    {
     k += (diag[i].BlockRowSet)->n;
    } /* end for i */
  printf("Check on set diag[].BlockRowSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=ndiag;i++)
    {
     k += (diag[i].NextLxkSet)->n;
    } /* end for i */
  printf("Check on set diag[].NextLxkSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=ndiag;i++)
    {
     k += (diag[i].NextLBxkSet)->n;
    } /* end for i */
  printf("Check on set diag[].NextLBxkSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=nlborder;i++)
    {
     k += (lborder[i].BorderRowSet)->n;
    } /* end for i */
  printf("Check on set lborder[].BorderRowSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

} /* end BDBFormData() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void OutputBDBFormData(FileNameA,FileNameb,FileNameN2,nprocs)
  char *FileNameA,*FileNameb,*FileNameN2;
  int nprocs;
{
/* output file format for block-bordered diagonal A matrix

(nblocks - dimension of blocks[])
blocks[1].BlockID       blocks[1].BlockSz       blocks[1].DiagPtr
     blocks[1].LBPtr 
.
.
.
blocks[nblocks].BlockID blocks[nblocks].BlockSz blocks[nblocks].DiagPtr
     blocks[nblocks].LBPtr

(n - dimension of diag[] -  i.e., the matrix dimension)
diag[1].Value diag[1].PtrLix diag[1].NextLxk diag[1].NextLBxk
.
.
.
diag[n].Value diag[n].PtrLix diag[n].NextLxk diag[n].NextLBxk

(nblockrow - dimension of blockrow[])
blockrow[1].Value         blockrow[1].i         blockrow[1].j
     blockrow[1].NextLxk
.
.
.
blockrow[nblockrow].Value blockrow[nblockrow].i blockrow[nblockrow].j
     blockrow[nblockrow].NextLxk

(nlborder - dimension of lborder[])
lborder[1].i        lborder[1].PtrLBix
.
.
.
lborder[nlborder].i lborder[nlborder].PtrLBix

(nborderrow - dimension of borderrow[])
borderrow[1].Value          borderrow[1].i          borderrow[1].j
     borderrow[1].NextLBxk
.
.
.
borderrow[nborderrow].Value borderrow[nborderrow].i borderrow[nborderrow].j
     borderrow[nborderrow].NextLBxk

ln lnnz
1 1 la(1,1)
.
.
.
i j la[i,j]
.
.
.
ln ln la(ln,ln)

EOF_input_BBD_matrix

Input file format for the b vector

b[1]
.
.
.
b[n]

EOF_input_b_vector

*/

 int i,j,ndb;
 int nlastblockstart,nlastblockend;
 FILE *fd;
printf("FileNameA =|%s| \n",FileNameA);
printf("FileNameb =|%s| \n",FileNameb);
fflush(stdout);
 fd = fopen(FileNameA,"w");
 
printf("write out the assignment of contigious diagonal elements to a processor nprocs=%d \n",nprocs);
fflush(stdout);
 
/* (0) write out the assignment of contigious diagonal elements to a processor */
  fprintf(fd," \t%d\n",nprocs);
  for (i=1; i<=nprocs; i++)
   {
    fprintf(fd,"%d \t%d %d  %d  %d  %d  %d \n",i,(i-1),
       proc[i].BlockPtr,proc[i].DiagPtr,proc[i].BlockRowPtr,proc[i].LowerBorderPtr,proc[i].BorderRowPtr);
   } /* end for i */ 
 
printf("write out the block identifer array of structures \n");
fflush(stdout);
 
/* (1) write out the block identifer array of structures */
  fprintf(fd," \t%d\n",nblocks);
  for (i=1; i<=nblocks; i++)
   {
    fprintf(fd,"%d \t%d %d %d %d %d \n",i,blocks[i].BlockID,blocks[i].BlockSz,
            blocks[i].ProcNum,blocks[i].DiagPtr,blocks[i].LBPtr );
   } /* end for i */
 
printf("write out the matrix diagonal array of structures \n");
fflush(stdout);
 
/* (2) write out the matrix diagonal array of structures */
  ndb = ndiag - (N2->n);
  fprintf(fd," \t%d\n",ndb);
  for (i=1; i<=ndb; i++)
   {
    fprintf(fd,"%d \t%lf %d %d %d \n",i,diag[i].Value,diag[i].PtrLix,
            diag[i].NextLxk,diag[i].NextLBxk);
   } /* end for i */
 
printf("write out the block row storage array of structures \n");
fflush(stdout);
 
/* (3) write out the block row storage array of structures */
  fprintf(fd," \t%d\n",nblockrow);
  for (i=1; i<=nblockrow; i++)
   {
    fprintf(fd,"%d \t%lf %d %d %d \n",i,blockrow[i].Value,blockrow[i].i,
            blockrow[i].j,blockrow[i].NextLxk );
   } /* end for i */
 
printf("write out the lower border row storage array of structures \n");
fflush(stdout);
 
/* (5) write out the lower border row storage array of structures */
  fprintf(fd," \t%d\n",nlborder);
  for (i=1; i<=nlborder; i++)
   {
    fprintf(fd,"%d \t%d %d\n",i,lborder[i].i,lborder[i].PtrLBix );
   } /* end for i */
 
printf("write out the border row storage array of structures \n");
fflush(stdout);
 
/* (7) write out the border row storage array of structures */
  fprintf(fd," \t%d\n",nborderrow);
  for (i=1; i<=nborderrow; i++)
   {
    fprintf(fd,"%d \t%lf %d %d %d\n",i,borderrow[i].Value,borderrow[i].i,
            borderrow[i].j,borderrow[i].NextLBxk );
   } /* end for i */
 
printf("write-out the last block \n");
fflush(stdout);
 
/* write-out the last block */
  fprintf(fd," \t%d %d\n",ln,lnnz);
  for (i=1; i<=lnnz; i++)
   {
    fprintf(fd,"%d \t%d %d %lf \n",i,lastbk[i].i,lastbk[i].j,lastbk[i].Value);
   } /* end for i */
 
printf("write out a character string to identify EOF \n");
fflush(stdout);
 
/* (9) write out a character string to identify EOF */
  fprintf(fd,"\nEOF_input_BBDmatrix\n");

  fclose(fd);

  fd = fopen(FileNameb,"w");
  if (fd == NULL)
   {
    printf("can't open file %s\n",FileNameb);
    exit(1);
   } /* end if */
 
printf("write out the b vector \n");
fflush(stdout);
 
/* (10) write out the b vector */
  for (i=1; i<=n; i++)
   {
    fprintf(fd,"%lf\t",(2.0*drand48()));
    if( (i%5) == 0) fprintf(fd,"\n");
   } /* end for i */
 
/* (11) write out a character string to identify EOF */
  fprintf(fd,"\n\nEOF_input_b_vector\n");

 fclose(fd);

/* write out the edges in the last block */

printf("write out the edges in the last block --- PrintN2=%d \n",PrintN2);
fflush(stdout);

  if (PrintN2 == YES)
   {
    fd = fopen(FileNameN2,"w");
    if (fd == NULL)
     {
      printf("can't open file %s\n",FileNameN2);
      exit(1);
     } /* end if */
    nlastblockstart = n - (N2->n) + 1;
    nlastblockend   = n;
    fprintf(fd,"%d %d %d \n",(N2->n),nlastblockstart,nlastblockend);
    j = 1;
    for (i=1;i<=nadjlist;i++)
      {
        if ( (adjlist[i].l.NodeNum >= nlastblockstart) && 
             (adjlist[i].r.NodeNum >= nlastblockstart))
	  {
           fprintf(fd,"%d \t%d %d \n",j++,adjlist[i].l.NodeNum,adjlist[i].r.NodeNum);
	  } /* end if */
      } /* end for i */
    fclose(fd);
    PrintN2 = NO;
   } /* end if PrintN2 == YES */

} /* end OutputBDBFormData() */

/* ****************************************************************************** */
/*                                                                                */
/*  Function:  X-Image(image-index,size,size,title)                                     */
/*                                                                                */
/*  derived fron an X-window display program developed by M. L. Katz of the MITRE */
/*  Corporation.  The original program was called image256.c.                     */
/*                                                                                */
/*  Description:  This program takes a matrix containing bytes that               */
/*                represent a digital image and displays that image               */
/*                in an X11 window.                                               */
/*                                                                                */
/* ****************************************************************************** */

void Ximage(xgs,ygs,Title)
int xgs,ygs;
char *Title;
{
    Window win;
    int x = 0, y = 0;           /* window position */
    unsigned int border_width = 4;  /* border four pixels wide */
    char *window_name = Title;
    char *icon_name = "gray";
    Pixmap icon_pixmap;
    XSizeHints size_hints;
    XEvent report;
    GC gc;
    char *display_name = NULL;
    int window_size = 0;  /* OK, or too SMALL to display contents */
/*
printf("width=|%d| - height=|%d| - title=|%s| \n",xgs,ygs,Title);
fflush(stdout);	
*/
    /* 1) set window width and height as specified by the arguments */
    width  = xgs;
    height = ygs;
/*
printf("connect to the server \n");
fflush(stdout);
*/
    /* 2) connect to the server */
    if ( (display = XOpenDisplay(display_name)) == NULL )
    {
    (void) fprintf( stderr,
        "gray scale: cannot connect to X server %s\n",
        XDisplayName(display_name));
    exit(-1);
    }
/*
printf("set the screen \n");
fflush(stdout);
*/
    /* 3) set the screen */
    /* get screen size from display structure macro */
    screen = DefaultScreen(display);
/*
printf("create a window \n");
fflush(stdout);
*/
    /* 4) create a window */
    win = XCreateSimpleWindow(display, RootWindow(display,screen),
        x, y, width, height, border_width,
        BlackPixel(display, screen),
        WhitePixel(display, screen));
/*
printf("load the image data from the image matrix \n");
fflush(stdout);
*/
    /* 10) load the image data from the image matrix */
    load_data();
/*
printf("prepare an icon pixmap \n");
fflush(stdout);
*/
    /* 5) prepare an icon pixmap */
    icon_pixmap = XCreateBitmapFromData(display, win, icon_bitmap_bits,
    icon_bitmap_width, icon_bitmap_height);

    /* initialize size hint property for window manager */
    size_hints.flags = PPosition | PSize | PMinSize;
    size_hints.x = x;
    size_hints.y = y;
    size_hints.width = width;
    size_hints.height = height;
    size_hints.min_width = width;
    size_hints.min_height = height;
/*
printf("set standard window properties \n");
fflush(stdout);
*/
    /* 6) set standard window properties */
    XSetStandardProperties(display, win, window_name, icon_name,
    icon_pixmap, gargv, gargc, &size_hints);
/*
printf("select event types \n");
fflush(stdout);
*/
    /* 7) select event types */
    XSelectInput(display, win, ExposureMask | ButtonPressMask |
    StructureNotifyMask);
/*
printf("define a gray scale colormap \n");
fflush(stdout);
*/
    /* 8) define a gray scale colormap */
    set_gray_scale(win);
/*
printf("create a graphics context \n");
fflush(stdout);
*/
    /* 9) create a graphics context */
    get_GC(win, &gc);
/*
printf("display the window \n");
fflush(stdout);
*/
    /* 11) display the window */
    XMapWindow(display, win);
/*
printf("draw the image in the window\n");
fflush(stdout);
*/
    /* 12) draw the image in the window */
    XPutImage(display, win, gc, picture, 0, 0, 0, 0, width, height);
/*
printf("process events \n");
fflush(stdout);
*/
    /* 13) process events */
    /* get events, use first Expose to display graphics;
     * ConfigureNotify to iniate a resize;  ButtonPress to exit
     */
    while(1)
    {
    XNextEvent(display, &report);
    switch (report.type)
    {
        case Expose:
        /* get rid of all other Expose events on the queue */
        while (XCheckTypedEvent(display, Expose, &report));
        if (window_size == SMALL)
            /* if window was resized too small to use */
           break;
        else
            /* place image in window */
            XPutImage(display, win, gc, picture, 0, 0, 0, 0, width, height);
        break;
        case ConfigureNotify:
        /* window has been resized, change width and
         * height to send to draw_gray_scale in next
         * Expose
         */
        width = report.xconfigure.width;
        height = report.xconfigure.height;
        if ((width < size_hints.min_width) ||
            (height < size_hints.min_height))
            window_size = SMALL;
        else
            window_size = OK;
        break;
        case ButtonPress:
        XFreeGC(display, gc);
        XCloseDisplay(display);
        return;
        default:
        /* all events selected by StructureNotifyMask
         * except ConfigureNotify are thrown away here.
         * since nothing is done with them.
         */
        break;
    }  /* end switch */
    }  /* end while */
}

/* ***************************************************************** */
/*                                                                   */
/*  This routine creates and installs a colormap with 256 grayscale  */
/*  colors.                                                          */
/*                                                                   */
/* ***************************************************************** */
void set_gray_scale(win)
Window win;
{
    Visual *visual;
    Colormap cmap;
    int  i;
    XColor cell_defs[MAX_COLORS];
    int white, black;

    /* set visual to default visual */
    visual = DefaultVisual(display, screen);

    /* create colormap with default visual and all color cells allocated */
    cmap = XCreateColormap(display, win, visual, AllocAll);

    /* install the new colormap for the given window */
    XSetWindowColormap(display, win, cmap);

    black = BlackPixel(display, screen);
    white = WhitePixel(display, screen);

    /* set white pixel as in standard colormap
    cell_defs[255].pixel = white;                         white
    cell_defs[255].red = 65535;
    cell_defs[255].green = 65535;
    cell_defs[255].blue = 65535;
    cell_defs[255].flags = DoRed | DoGreen | DoBlue;

     set black pixel as in standard colormap
    cell_defs[0].pixel = black;                         black
    cell_defs[0].red = 0;
    cell_defs[0].green = 0;
    cell_defs[0].blue = 0;
    cell_defs[0].flags = DoRed | DoGreen | DoBlue;
*/
    /* set color cells 2-254 to represent gray scale colors */
/*    for (i=1; i<(MAX_COLORS-2); i++) */
    for (i=0; i<(MAX_COLORS); i++)
    {
    cell_defs[i].pixel = i;
    cell_defs[i].red = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].green = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].blue = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].flags = DoRed | DoGreen | DoBlue;
    }

    /* store color cell colors in new colormap */
    XStoreColors(display, cmap, cell_defs, MAX_COLORS);
}

/* ******************************************************************* */
/*                                                                     */
/*  This routine creates a graphics context by calling the X Windows   */
/*  routine XCreateGC.  A default graphics context is created without  */
/*  specifying any of the attribute values.                            */
/*                                                                     */
/* ******************************************************************* */
void get_GC(win, gc)
Window win;
GC *gc;
{
    unsigned long valuemask = 0;  /* ignore XGCvalues and use defaults */
    XGCValues values;

    *gc = XCreateGC(display, win, valuemask, &values);
}


/* ***************************************************************** */
/*                                                                   */
/*  This routine loads the data file that represents the digital     */
/*  image.                                                           */
/*                                                                   */
/* ***************************************************************** */
int load_data()
{
    int i, j;
    unsigned char *pic;
    Visual *vis;
    int depth;

    /* allocate memory to store image */
    if ((pic = (unsigned char *) calloc(width, height, sizeof(unsigned char)))
                            == NULL)
    {
    sprintf(stderr, "Error allocating array\n");
    return(0);
    }
    /* adjust values in image to correspond to colormap */
      for (i = 0; i < height; i++)
    for (j = 0; j < width; j++)
          {
            *(pic+(i*width+j)) = bitmap[j+1][i+1];

        if (*(pic+(i*width+j)) == 255)
        *(pic+(i*width+j)) = 0;
        else
        (*(pic+(i*width+j)))++;
    }

    vis = DefaultVisual(display, screen);
    depth = DisplayPlanes(display, screen);

    /* create an image structure in X */
    picture = XCreateImage(display, vis, depth, ZPixmap, 0,
        pic, width, height, 8, width);
}


